import os
import subprocess
import time
import distutils.util
import socket
import time

from firexapp.common import qualify_firex_bin, select_env_vars
from firexapp.submit.submit import OptionalBoolean
from firexapp.submit.tracking_service import TrackingService
from firexapp.submit.console import setup_console_logging
from firexapp.submit.install_configs import FireXInstallConfigs

from firex_flame.flame_helper import DEFAULT_FLAME_TIMEOUT, get_flame_debug_dir, get_rec_file, is_json_file, \
    get_flame_redirect_file_path
from firex_flame.model_dumper import is_dump_complete, get_run_metadata_file, wait_and_get_flame_url
from firexapp.discovery import PkgVersionInfo

logger = setup_console_logging(__name__)


def _send_flame_indicate_ready(celery_app):
    if celery_app:
        with celery_app.events.default_dispatcher(hostname=socket.gethostname()) as d:
            # Flame will indicate it's ready for tasks once it gets this Celery event.
            return d.send('flame-indicate-ready')
    else:
        logger.warning(
            "Flame launcher did not receive Celery app, "
            "cannot send event for server to indicate readiness.")


def get_flame_args(uid, broker_recv_ready_file, args):
    if args.flame_record:
        rec_file = args.flame_record
    else:
        rec_file = get_rec_file(uid.logs_dir)

    # assemble startup cmd
    cmd_args = {
        'port': args.flame_port,
        'uid': str(uid),
        'logs_dir': uid.logs_dir,
        'chain': args.chain,
        'recording': rec_file,
        'central_server': args.flame_central_server,
        'central_server_ui_path': args.flame_central_server_ui_path,
        'logs_server': args.flame_logs_server,
        'central_documentation_url': args.flame_central_documentation_url,
        'flame_timeout': args.flame_timeout,
        'broker_recv_ready_file': broker_recv_ready_file,
        'broker_max_retry_attempts': args.broker_max_retry_attempts,
        'terminate_on_complete': args.flame_terminate_on_complete,
        'firex_bin_path': args.firex_bin_path,
        'extra_task_dump_paths': args.flame_extra_task_dump_paths,
        'serve_logs_dir': args.flame_serve_logs_dir,
        'authed_user_request_path': args.flame_authed_user_request_path,
        'wait_for_webserver': args.flame_wait_for_webserver,
    }
    result = []
    for k, v in cmd_args.items():
        if v is not None:
            result.append(f'--{k}')
            result.append('%s' % v)
    return result


class FlameLauncher(TrackingService):

    def __init__(self):
        self.broker_recv_ready_file = None
        self.sync = None
        self.firex_logs_dir = None
        self.is_ready_for_tasks = False
        self.start_time = None
        self.stdout_file = None
        self.wait_for_webserver = None

    def extra_cli_arguments(self, arg_parser):
        arg_parser.add_argument('--flame_timeout', help='How long the webserver should run for, in seconds.',
                                default=DEFAULT_FLAME_TIMEOUT)
        arg_parser.add_argument('--flame_central_server',
                                help='Server URL from which flame resources can be fetched to enable browser caching'
                                     'and client-side settings.',
                                default=None)
        arg_parser.add_argument('--flame_central_server_ui_path',
                                help='Path relative to flame_central_server from which the Flame UI is served.',
                                default=None)
        arg_parser.add_argument('--flame_logs_server',
                                help='Server URL from which flame logs can be fetched.',
                                default=None)
        arg_parser.add_argument('--flame_central_documentation_url',
                                help='URL linking to main out-of-app docs.',
                                default=None)
        arg_parser.add_argument('--firex_bin_path',
                                help='Path to firex executable.',
                                default=None)
        arg_parser.add_argument('--broker_max_retry_attempts',
                                help='See Flame argument help.',
                                default=None)
        arg_parser.add_argument('--flame_record',
                                help='A file to record flame events',
                                default=None)
        arg_parser.add_argument('--flame_port',
                                help='Flame port to be used', type=int,
                                default=0)
        arg_parser.add_argument('--flame_terminate_on_complete',
                                help='Terminate Flame when run completes. Ignores timeout arg entirely.',
                                default=None, const=True, nargs='?', action=OptionalBoolean)
        arg_parser.add_argument('--flame_wait_for_webserver',
                                help='Wait for webserver when waiting to be ready for tasks.',
                                default=True, const=True, nargs='?', action=OptionalBoolean)
        arg_parser.add_argument('--flame_extra_task_dump_paths',
                                help='Paths specifying alternative task represetnation to dump at end of flame.',
                                default=None)
        arg_parser.add_argument('--flame_authed_user_request_path', default=None)
        arg_parser.add_argument('--flame_serve_logs_dir',
                                help="Control if the Flame server makes the run's logs_dir available via HTTP(S).",
                                type=lambda x: bool(distutils.util.strtobool(x)),
                                default=None)


    def start(self, args, install_configs: FireXInstallConfigs, uid=None, **kwargs) -> dict:
        super().start(args, install_configs, uid=uid, **kwargs)

        flame_debug_dir = get_flame_debug_dir(uid.logs_dir)
        os.makedirs(flame_debug_dir, exist_ok=True)
        self.broker_recv_ready_file = os.path.join(flame_debug_dir, 'celery_receiver_ready')

        self.sync = args.sync
        self.firex_logs_dir = uid.logs_dir

        self.wait_for_webserver = args.flame_wait_for_webserver

        flame_args = get_flame_args(uid, self.broker_recv_ready_file, args)
        self.stdout_file = os.path.join(flame_debug_dir, 'flame.stdout')

        self.start_time = time.time()
        try:
            with open(self.stdout_file, 'w+') as f:
                subprocess.Popen([qualify_firex_bin("firex_flame")] + flame_args,
                                 stdout=f, stderr=subprocess.STDOUT,
                                 close_fds=True,
                                 env=select_env_vars(['PATH']),
                                 preexec_fn=os.setpgrp, # Avoid SIGINTs sent to FireX by having own process group.
                                 cwd=flame_debug_dir,
                                 )
        except Exception as e:
            logger.error("Flame subprocess start failed: %s." % e)
            raise

        return {}

    def ready_for_tasks(self, celery_app=None, **kwargs) -> bool:
        if not self.is_ready_for_tasks:
            # Flame will receive this event once its receiving from Celery
            # and create broker_recv_ready_file
            _send_flame_indicate_ready(celery_app)
            time.sleep(0.1)
            broker_ready = os.path.isfile(self.broker_recv_ready_file)

            if self.wait_for_webserver:
                run_metadata_file = get_run_metadata_file(firex_logs_dir=self.firex_logs_dir)
                webserver_ready = os.path.isfile(run_metadata_file) and is_json_file(run_metadata_file)
            else:
                webserver_ready = True

            self.is_ready_for_tasks = broker_ready and webserver_ready
            if self.is_ready_for_tasks:
                logger.debug("Flame up after %.2f s" % (time.time() - self.start_time))
                if self.wait_for_webserver:
                    # Only print Flame URL if we've waited for webserver, since otherwise we don't know the port.
                    logger.info(f"Flame: {self.get_viewer_url()}")
                self.write_flame_redirect()

        return self.is_ready_for_tasks

    def write_flame_redirect(self):
        # Write the flame redirect file in the run's logs dir
        flame_redirect_filepath = get_flame_redirect_file_path(self.firex_logs_dir)
        with open(flame_redirect_filepath, 'w') as f:
            f.write(f'<meta http-equiv="refresh" content="0; url={self.get_viewer_url()}" />')

    def ready_release_console(self, **kwargs) -> bool:
        if self.sync:
            # For sync requests, guarantee that the model is completely dumped before terminating.
            return is_dump_complete(self.firex_logs_dir)
        return True

    def get_viewer_url(self):
        if (
            not self.install_configs.has_viewer()
            and self.firex_logs_dir
        ):
            return wait_and_get_flame_url(self.firex_logs_dir, timeout=0)
        return self.install_configs.run_url

    @staticmethod
    def get_pkg_version_info() -> PkgVersionInfo:
        import firex_flame
        return PkgVersionInfo(pkg='firex-flame',
                              version=firex_flame.__version__,
                              commit=firex_flame._version.get_versions()['full-revisionid'])


import logging
from typing import Optional, Any, Union
import re
from datetime import datetime
import dataclasses
import os
import tarfile
from pathlib import Path
from enum import Enum
import json
import time
import shutil
import subprocess

from firexapp.events.model import ADDITIONAL_CHILDREN_KEY, EXTERNAL_COMMANDS_KEY, RunStates
from firexapp.events.event_aggregator import transform_task_state
from firex_flame.flame_helper import flatten, deep_merge
from firex_flame.model_dumper import get_all_tasks_dir, get_tasks_slim_file, get_run_metadata_file, \
    get_model_complete_file, atomic_write_json, get_flame_model_dir

from gevent.lock import BoundedSemaphore


logger = logging.getLogger(__name__)

LIST_PATH_ENTRY = re.compile(r'^\[(\d+)\]$')

TASK_TYPE = dict[str, Any] # fixme should probably data model
TASKS_BY_UUID_TYPE = dict[str, TASK_TYPE] # fixme should probably data model

TASK_ARGS = 'firex_bound_args'


def _times_from_event(event: dict[str, Any]) -> dict:
    times = dict(latest_timestamp=event['local_received'])
    if event.get('type') == 'task-started-info':
        # Note first_started is never overwritten by aggregation.
        times['first_started'] = event['local_received']
    return times


# config field options:
#   copy_celery - True if this field should be copied from the celery event to the task data model. If the field already
#                   has a value on the data model, more recent celery field values will overwrite existing values by
#                   default. If overwriting should be avoided, see 'aggregate_merge' and 'aggregate_keep_initial'
#                   options described below.
#
#   slim_field - True if this field should be included in the 'slim' (minimal) data model representation sent to the UI.
#                   Be very careful adding fields, since this field will be sent for each node and can therefore greatly
#                   increase the amount of data sent to the UI on main graph load.
#
#   transform_celery - A function to be executed on the entire event when the corresponding key is present in a celery
#                       event. The function returns a dict that dict.update the existing data from the event, possibly
#                       overwriting data copied from the celery event by copy_celery=True. Can be used to change
#                       the field name on the data model from the field name from celery.
#
#   aggregate_merge - True if model updates should deep merge collection data types (lists, dicts, sets) instead of
#                       overwriting.
#
#   aggregate_keep_initial - True if data model field updates should be ignored after an initial value has been set.
#                               This is one way of preventing overwriting, see also 'aggregate_merge'.
#
#
FIELD_CONFIG = {
    'uuid': {'copy_celery': True, 'slim_field': True},
    'hostname': {'copy_celery': True, 'slim_field': True},
    'parent_id': {'copy_celery': True, 'slim_field': True},
    'type': {
        'copy_celery': True,
        'transform_celery': transform_task_state,
    },
    'retries': {'copy_celery': True, 'slim_field': True},
    TASK_ARGS: {'copy_celery': True},
    'actual_runtime': {'copy_celery': True, 'slim_field': True},
    'support_location': {'copy_celery': True},
    'utcoffset': {'copy_celery': True},
    'code_url': {'copy_celery': True},
    # TODO: start using code_filepath instead of code_url.
    'code_filepath': {'copy_celery': True},
    'firex_default_bound_args': {'copy_celery': True},
    'from_plugin': {'copy_celery': True, 'slim_field': True},
    'chain_depth': {'copy_celery': True, 'slim_field': True},
    'firex_result': {'copy_celery': True},
    'traceback': {'copy_celery': True},
    'exception': {'copy_celery': True},
    'long_name': {
        'copy_celery': True,
        'transform_celery': lambda e: {'name': e['long_name'].split('.')[-1]},
    },
    'flame_data': {
        'copy_celery': True,
        'slim_field': True,
        'aggregate_merge': True,
    },
    'state': {'slim_field': True},
    'task_num': {'slim_field': True},
    'name': {
        'slim_field': True,
        # TODO: firexapp should send long_name, since it will overwrite 'name' copied from celery. Then get rid of
        # the following config.
        'transform_celery': lambda e: {
            'name': e['name'].split('.')[-1],
            'long_name': e['name']},
    },
    'called_as_orig': {
        'copy_celery': True,
    },
    'first_started': {
        'slim_field': True,
        'aggregate_keep_initial': True,
    },
    'url': {
        # TODO: only for backwards compat. Can use log_filepath.
        'transform_celery': lambda e: {'logs_url': e['url']},
    },
    'log_filepath': {
        'transform_celery': lambda e: {'logs_url': e['log_filepath']},
    },
    'local_received': {
        'transform_celery': _times_from_event,
    },
    'states': {'aggregate_merge': True},
    'exception_cause_uuid': {
        'copy_celery': True,
        'slim_field': True,
    },
    EXTERNAL_COMMANDS_KEY: {
        'copy_celery': True,
        'aggregate_merge': True,
    },
    ADDITIONAL_CHILDREN_KEY: {
        'copy_celery': True,
        'aggregate_merge': True,
        'slim_field': True,
    },
    'error_context': {
        'copy_celery': True,
        'slim_field': True,
    },
    'cached_result_from': {
        'copy_celery': True,
    },
    'latest_timestamp': {
        'slim_field': True,
    },
    'pid': {
        'copy_celery': True,
    },
    'was_revoked': {
        'copy_celery': True,
    },
}


def _get_keys_with_true(input_dict, key):
    return [k for k, v in input_dict.items() if v.get(key, False)]


COPY_FIELDS = _get_keys_with_true(FIELD_CONFIG, 'copy_celery')
# These are the minimum fields required to render the graph.
SLIM_FIELDS = _get_keys_with_true(FIELD_CONFIG, 'slim_field')

AGGREGATE_MERGE_FIELDS = _get_keys_with_true(FIELD_CONFIG, 'aggregate_merge')
AGGREGATE_KEEP_INITIAL_FIELDS = _get_keys_with_true(FIELD_CONFIG, 'aggregate_keep_initial')
AGGREGATE_NO_OVERWRITE_FIELDS = AGGREGATE_MERGE_FIELDS + AGGREGATE_KEEP_INITIAL_FIELDS

FIELD_TO_CELERY_TRANSFORMS = {
    k: v['transform_celery'] for k, v in FIELD_CONFIG.items()
    if 'transform_celery' in v}

class _TaskFieldSentile(Enum):
    UNSET = 1
    UNLOADED = 2

@dataclasses.dataclass
class _ModelledFlameTask:
    firex_bound_args: Union[dict, _TaskFieldSentile] = _TaskFieldSentile.UNSET
    firex_default_bound_args: Union[dict, _TaskFieldSentile] = _TaskFieldSentile.UNSET
    firex_result: Union[dict, _TaskFieldSentile] = _TaskFieldSentile.UNSET
    external_commands: Union[dict, _TaskFieldSentile] = _TaskFieldSentile.UNSET

    def as_dict(self):
        return {
            field_name: getattr(self, field_name)
            for field_name in self.unloadable_field_names()
            if not isinstance(getattr(self, field_name), _TaskFieldSentile)
        }

    def get_set_field_names(self):
        return [
            field_name for field_name in self.unloadable_field_names()
            if getattr(self, field_name) != _TaskFieldSentile.UNSET
        ]

    def unloadable_field_names(self):
        return [f.name for f in dataclasses.fields(self)]

    def any_unloaded(self, field_names=None) -> bool:
        if field_names is None:
            field_names = self.unloadable_field_names()
        return any(
            getattr(self, f) == _TaskFieldSentile.UNLOADED
            for f in field_names
        )

    def unload_fields(self):
        for field_name in self.unloadable_field_names():
            field_val = getattr(self, field_name)
            if not isinstance(field_val, _TaskFieldSentile):
                # allow this field's value to be garbage collected by unreferencing it.
                setattr(self, field_name, _TaskFieldSentile.UNLOADED)
                del field_val

    def update(self, update_dict):
        for k in self.unloadable_field_names():
            if k in update_dict:
                setattr(self, k, update_dict[k])


def is_task_dict_complete(task_dict) -> bool:
    return RunStates.is_complete_state(
        task_dict.get('state'),
        has_completed=task_dict.get('has_completed'),
    )


@dataclasses.dataclass
class _FlameTask:
    _modelled : _ModelledFlameTask
    always_loaded_task_data: dict[str, Any]
    model_dumper : 'FlameModelDumper'
    _lock = BoundedSemaphore()

    @staticmethod
    def create_task(task_uuid, task_num, model_dumper : 'FlameModelDumper'):
        return _FlameTask(
            _ModelledFlameTask(),
            always_loaded_task_data={
                'uuid': task_uuid,
                'state': None,
                'task_num': task_num,
                'was_revoked': False,
                # Indicates the completed event has been received,
                # DOES NOT GUARANTEE ALL TASK DATA HAS BEEN RECEIVED.
                # task-success (with results) can be received after task-completed,
                # for example.
                'has_completed': False,
            },
            model_dumper=model_dumper
        )

    def as_dict(self) -> dict[str, Any]:
        return self.always_loaded_task_data | self._modelled.as_dict()

    def _load_from_full_task_file(self):
        assert self._lock.locked()
        try:
            return self.model_dumper.load_full_task(self.get_uuid())
        except OSError as e:
            logger.warning(f"Failed to load full task for {self.get_uuid()} due to: {e}")
            return {}

    def _already_locked_get_full_task_dict(self, field_names=None):
        assert self._lock.locked()
        if self._modelled.any_unloaded(field_names):
            modeled_dict = self._load_from_full_task_file()
        else:
            modeled_dict = self._modelled.as_dict()
        return modeled_dict | self.always_loaded_task_data

    def get_full_task_dict(self):
        with self._lock:
            return self._already_locked_get_full_task_dict()

    def get_uuid(self):
        return self.always_loaded_task_data['uuid']

    def dump_full(self, new_event_types: set[str]):
        should_dump = (
            # dump when we receive args
            'task-started-info' in new_event_types

            # dump when the task is complete (or if it was already complete, regardless of event type, due to out-of-order events.)
            # FIXME: is_completed can cause double dumps for successes when completed is received after success.
            or self.is_complete()
        )
        if should_dump:
            with self._lock:
                full_task_dict = self._already_locked_get_full_task_dict()
                try:
                    self.model_dumper.dump_full_task(self.get_uuid(), full_task_dict)
                except OSError as e:
                    logger.warning(f"Failed to write {self.get_uuid()} full task JSON: {e}")
                else:
                    if self.is_complete():
                        # TODO: is it too much to always unload completed?
                        self._modelled.unload_fields()
                        task_num = full_task_dict['task_num']
                        if task_num % 100 == 0:
                            logger.debug(f'Unloaded big fields for task num {task_num}, uuid {self.get_uuid()}')

    def get_task_state(self):
        return self.always_loaded_task_data['state']

    def is_complete(self) -> bool:
        return is_task_dict_complete(self.always_loaded_task_data)

    def get_field(self, field_name, default=_TaskFieldSentile.UNSET):
        return self.get_fields([field_name]).get(field_name, default)

    def get_fields(self, field_names):
        if set(field_names).isdisjoint(self._modelled.unloadable_field_names()):
            task_dict = self.always_loaded_task_data
        else:
            with self._lock:
                task_dict = self._already_locked_get_full_task_dict(field_names)

        return {
            k: task_dict[k]
            for k in field_names
            if k in task_dict
        }

    def update(self, update_dict: dict[str, Any]):
        self._modelled.update(
            {
                k: v for k, v in update_dict.items()
                if k in self._modelled.unloadable_field_names()
            }
        )
        self.always_loaded_task_data.update(
            {
                k: v for k, v in update_dict.items()
                if k not in self._modelled.unloadable_field_names()
            }
        )

    def get_field_names(self):
        return list(self.always_loaded_task_data.keys()) + self._modelled.get_set_field_names()


class FlameTaskGraph:

    def __init__(
        self,
        tasks_by_uuid: Optional[TASKS_BY_UUID_TYPE] = None,
        model_dumper: Optional['FlameModelDumper'] = None,
    ):
        self.root_uuid : Optional[str] = None

        self._tasks_by_uuid : dict[str, _FlameTask] = {}
        self.model_dumper = model_dumper
        self._event_aggregator = FlameEventAggregator(
            self._tasks_by_uuid, # Must share instance.
            self.model_dumper)

        # cache relationships for faster querying for ancestors and descedants.
        self._parent_to_children_uuids: dict[str, set[str]] = {}
        self._child_to_parents_uuids: dict[str, set[str]] = {}

        # add input task to graph's instance.
        self.update_graph_from_celery_events(
            (tasks_by_uuid or {}).values()
        )

    def get_root_task(self) -> Optional[TASK_TYPE]:
        if not self.root_uuid:
            return None
        return self.get_full_task_dict(self.root_uuid)

    def _walk_task_graph(self, task_uuid, uuids_by_uuid) -> list[_FlameTask]:
        checked_uuids = set()
        to_check_uuids = [task_uuid]
        result_tasks_by_uuid = {}
        while to_check_uuids:
            uuid = to_check_uuids.pop()
            checked_uuids.add(uuid)
            for related_uuid in uuids_by_uuid.get(uuid, []):
                if related_uuid in self._tasks_by_uuid:
                    result_tasks_by_uuid[related_uuid] = self._tasks_by_uuid[related_uuid]
                if related_uuid not in checked_uuids:
                    to_check_uuids.append(related_uuid)
        return list(result_tasks_by_uuid.values())

    def get_descendants_of_uuid(self, task_uuid) -> list[_FlameTask]:
        return self._walk_task_graph(
            task_uuid,
            self._parent_to_children_uuids,
        )

    def get_ancestors_of_uuid(self, task_uuid) -> list[_FlameTask]:
        return self._walk_task_graph(
            task_uuid,
            self._child_to_parents_uuids,
        )

    def _add_parent_and_child(self, parent_uuid, child_uuid):
        if parent_uuid not in self._parent_to_children_uuids:
            self._parent_to_children_uuids[parent_uuid] = set()
        self._parent_to_children_uuids[parent_uuid].add(child_uuid)

        if child_uuid not in self._child_to_parents_uuids:
            self._child_to_parents_uuids[child_uuid] = set()
        self._child_to_parents_uuids[child_uuid].add(parent_uuid)

    def _maybe_set_root_uuid(self, events):
        for event in events:
            if self.root_uuid is None:
                if (
                    event.get('parent_id', '__no_match') is None
                    and event.get('uuid')
                ):
                    self.root_uuid = event['uuid']
                elif event.get('root_id') is not None:
                    # we can still know the root if we miss the first event (the root's event with parent_id)
                    # since other events reference the root UUID via root_id.
                    self.root_uuid = event['root_id']

    def _update_graph_from_task_data(self, task_uuid, event):
        if task_uuid:
            parent_id = event.get('parent_id')
            if parent_id:
                self._add_parent_and_child(parent_id, task_uuid)

            additional_children = event.get(ADDITIONAL_CHILDREN_KEY)
            if additional_children:
                for child_uuid in additional_children:
                    self._add_parent_and_child(task_uuid, child_uuid)

    def update_graph_from_celery_events(self, events):
        self._maybe_set_root_uuid(events)

        # find only data changes from events.
        new_data_by_task_uuid = self._event_aggregator.aggregate_events(events)
        for task_uuid, new_task_data in new_data_by_task_uuid.items():
            self._update_graph_from_task_data(task_uuid, new_task_data)

        slim_update_data_by_uuid = {
            uuid: task
            for uuid, task in _slim_tasks_by_uuid(new_data_by_task_uuid).items()
            if task
        }
        return new_data_by_task_uuid, slim_update_data_by_uuid

    def query_partial_tasks(self, query_task_uuids, task_queries):
        # When querying a partial set of tasks, count descendants as matches to be included in the result.
        return _query_flame_tasks(
            self,
            query_task_uuids,
            task_queries,
            match_descendant_criteria=True)

    def query_full_tasks(self, task_queries):
        # When querying a full set of tasks, descendants will be included when their ancestors are matched.
        return _query_flame_tasks(
            self,
            self.get_all_task_uuids(),
            task_queries,
            match_descendant_criteria=False)

    def is_root_started(self) -> bool:
        task = self._tasks_by_uuid.get(self.root_uuid)
        if task is None:
            return False
        return task.get_task_state() not in [None, RunStates.RECEIVED.to_celery_event_type()]

    def is_root_complete(self) -> bool:
        task = self._tasks_by_uuid.get(self.root_uuid)
        if task is None:
            return False
        return task.is_complete()

    def all_tasks_complete(self):
        if not self._tasks_by_uuid:
            # do not want "no tasks received yet"
            # to look like "all tasks complete"
            return False

        # In case checking all tasks is expensive, check root first. Remaining
        # tasks only need to be checked once root is complete since everything
        # can't be complete if the root is not complete.
        if not self.is_root_complete():
            return False

        return all(
            t.is_complete() for t in self._tasks_by_uuid.values()
        )

    def get_full_task_dict(self, uuid) -> Optional[dict[str, dict[str, Any]]]:
        maybe_full_task = self._tasks_by_uuid.get(uuid)
        if maybe_full_task is None:
            return None
        return maybe_full_task.get_full_task_dict()

    def get_slim_task_dict(self, uuid) -> Optional[dict[str, dict[str, Any]]]:
        task = self._tasks_by_uuid.get(uuid)
        if task is None:
            return None
        return task.get_fields(SLIM_FIELDS)

    def get_task_field(self, uuid, field_name, default=None):
        task = self._tasks_by_uuid.get(uuid)
        if task is None:
            return default
        return task.get_field(field_name, default)

    def get_all_task_uuids(self):
        return set(self._tasks_by_uuid.keys())

    def get_slim_tasks_by_uuid(self):
        return self.get_all_tasks_fields(SLIM_FIELDS)

    def get_all_tasks_fields(self, fields):
        return {
            uuid: task.get_fields(fields)
            for uuid, task in self._tasks_by_uuid.items()
        }

    def get_full_tasks_by_uuid(self) -> TASKS_BY_UUID_TYPE:
        return {u: t.get_full_task_dict() for u, t in self._tasks_by_uuid.items()}

    def set_and_dump_any_incomplete_tasks(self):
        # Create new events that change the run state of incomplete events.
        incomplete_task_events = self._event_aggregator.generate_incomplete_events()
        if incomplete_task_events:
            logger.warning(f"Forcing runstates of {len(incomplete_task_events)} incomplete tasks to be terminal.")
            new_data_by_task_uuid, slim_update_data_by_uuid = self.update_graph_from_celery_events(incomplete_task_events)
        else:
            logger.debug("All tasks already terminal following terminal root.")
            new_data_by_task_uuid = slim_update_data_by_uuid = {}

        for uuid in new_data_by_task_uuid:
            self._tasks_by_uuid[uuid].dump_full(set())

        return new_data_by_task_uuid, slim_update_data_by_uuid

    def dump_full_task(self, uuid: str, new_event_types: set[str]):
        if uuid not in self._tasks_by_uuid:
            logger.warning(f'Ignoring request to dump non-existant tast {uuid}')
        else:
            self._tasks_by_uuid[uuid].dump_full(new_event_types)

    def _get_task(self, uuid) -> _FlameTask:
        # Everything outside this module should do get_full_task_dict instead of this.
        return self._tasks_by_uuid.get(uuid)

    def was_revoked(self, uuid) -> bool:
        task = self._get_task(uuid)
        if task is None:
            return False
        return task.get_field('was_revoked', False)

def _slim_tasks_by_uuid(tasks_by_uuid):
    return {uuid: {k: v for k, v in task_data.items()
                   if k in SLIM_FIELDS}
            for uuid, task_data in tasks_by_uuid.items()}

def _validate_task_queries(task_representation):
    if not isinstance(task_representation, list):
        return False

    missing_criterias = [r for r in task_representation
                         if 'matchCriteria' not in r
                            or not isinstance(r['matchCriteria'], dict)]
    if missing_criterias:
        return False

    # TODO: validate matchCriteria themselves

    return True


def _normalize_criteria_key(k):
    return k[1:] if k.startswith('?') else k


def _matches_equal_criteria(task: _FlameTask, eq_criteria: dict[str, Any]):
    # TODO: if more adjusting qualifiers are added, this needs to be reworked.
    required_keys = {k for k in eq_criteria.keys() if not k.startswith('?')}
    optional_keys = {_normalize_criteria_key(k) for k in eq_criteria.keys() if k.startswith('?')}

    task_fields = task.get_field_names()
    if not required_keys.issubset(task_fields):
        return False # task is missing required keys, can't match.

    queried_field_names = required_keys.union(optional_keys)
    normalized_criteria = {_normalize_criteria_key(k): v for k, v in eq_criteria.items()}
    for task_field in task_fields: # ordered to avoid loading unloadable fields
        if (
            task_field in queried_field_names
            and task.get_field(task_field) != normalized_criteria[task_field]
        ):
            return False
    return True

def _matches_has_key_criteria(task: _FlameTask, key_path):
    if isinstance(key_path, str):
        key_path = key_path.split('.')

    if key_path and isinstance(key_path, list):
        first_key = key_path[0]
        # try to avoid loading full task for in-memory keys instead
        # of getting full task dict.
        first_key_val = task.get_field(first_key)
        if first_key_val != _TaskFieldSentile.UNSET:
            remaining_keys = key_path[1:]
            if not remaining_keys:
                return True
            elif isinstance(first_key_val, dict):
                tmp_dict = first_key_val
                for k in remaining_keys:
                    try:
                        tmp_dict = tmp_dict[k]
                    except Exception:
                        return False
                return True
    return False


def task_matches_criteria(task: _FlameTask, criteria: dict):
    if criteria['type'] == 'all':
        return True

    if criteria['type'] == 'always-select-fields':
        # always-select-fields doesn't cause matches (tasks to be included), but paths here are always included
        # in results.
        return False

    if criteria['type'] == 'equals':
        return _matches_equal_criteria(task, criteria['value'])

    if criteria['type'] == 'has-key':
        return _matches_has_key_criteria(task, criteria['value'])

    return False


def _add_path_to_container(container, path_list, val):
    if not path_list:
        return
    if len(path_list) == 1:
        final_key = path_list[0]
        is_list = LIST_PATH_ENTRY.match(final_key)
        if is_list:
            container.append(val)
        else:
            container[final_key] = val
    else:
        cur_key = path_list.pop(0)
        is_cur_list = LIST_PATH_ENTRY.match(cur_key)
        if is_cur_list:
            cur_key = int(is_cur_list.group(1))

def _add_path_to_container(top_container: dict[str, Any], path_list: tuple[Union[str,int]], val):
    latest_container = top_container
    for i, cur_key in enumerate(path_list):
        is_last_key = i == len(path_list) - 1
        is_latest_list = isinstance(latest_container, list)

        if is_latest_list:
            latest_cont_keys = range(len(latest_container))
        else:
            latest_cont_keys = latest_container.keys()

        if is_last_key:
            # Last key, set value instead of finding next container.
            if is_latest_list and cur_key not in latest_cont_keys:
                latest_container.append(val)
            else:
                latest_container[cur_key] = val
        else:
            # not last key, find or create next container.
            if cur_key not in latest_cont_keys:
                is_next_list = isinstance(path_list[i+1], int)
                if is_next_list:
                    next_container = []
                else:
                    next_container = {}

                if is_latest_list:
                    latest_container.append(next_container) # assume integer keys pre-sorted.
                else:
                    latest_container[cur_key] = next_container
            else:
                next_container = latest_container[cur_key]
            latest_container = next_container


def _int_index_or_key(json_key_part_str):
    m = LIST_PATH_ENTRY.match(json_key_part_str)
    if m:
        return int(m.group(1))
    return json_key_part_str


def _container_from_json_paths_to_values(json_paths_to_values: list[str]):

    container = {}
    parsed_key_paths_to_values = {
        tuple(_int_index_or_key(str_k_part) for str_k_part in k.split('.')): v
        for k, v in json_paths_to_values.items()
    }
    for path_tuple, value in sorted(parsed_key_paths_to_values.items(), key=lambda kv: kv[0]):
        _add_path_to_container(
            container,
            path_tuple,
            value)
    return container


def _jsonpath_get_paths(jsonpath_exprs, task_dict):
    matching_paths_to_values = {}
    for jsonpath_expr in jsonpath_exprs:
        matching_paths_to_values.update(
            {
                str(match.full_path): match.value
                for match in jsonpath_expr.find(task_dict)
            }
        )
    return _container_from_json_paths_to_values(matching_paths_to_values)


def _get_descendants_for_criteria(select_paths, descendant_criteria, ancestor_uuid, task_graph: FlameTaskGraph):
    ancestor_descendants = task_graph.get_descendants_of_uuid(ancestor_uuid)
    matched_descendants_by_uuid = {}
    for criteria in descendant_criteria:
        for descendant in ancestor_descendants:
            if task_matches_criteria(descendant, criteria):
                # The fields that are selected for each descendant are determined by all queries, except
                # descendant descendants are never included.
                matched_descendants_by_uuid[descendant.get_uuid()] = select_from_task(
                    select_paths,
                    [],  # Never include descendants in descendant queries to avoid infinite loop.
                    descendant,
                    task_graph)

    return matched_descendants_by_uuid


def select_from_task(
    select_paths,
    select_descendants,
    task: _FlameTask,
    task_graph: FlameTaskGraph,
):
    selected_dict = {}
    # FIXME: get_full_task_dict may load here even when the paths don't need it to!
    paths_update_dict = _jsonpath_get_paths(select_paths, task.get_full_task_dict())
    selected_dict.update(paths_update_dict)

    selected_descendants_by_uuid = _get_descendants_for_criteria(
        select_paths,
        select_descendants,
        task.get_uuid(),
        task_graph)

    if selected_descendants_by_uuid:
        selected_dict.update({'descendants': selected_descendants_by_uuid})

    return selected_dict


def get_always_select_fields(task_queries):
    return flatten([q.get('selectPaths', []) for q in task_queries
                    if q['matchCriteria']['type'] == 'always-select-fields'])


def _select_ancestors_of_task_descendant_match(
    desc_task: _FlameTask,
    all_task_queries,
    task_graph: FlameTaskGraph,
):
    # Should the current task be included in the result because it matches some descendant criteria?
    ancestor_results_by_uuid = {}

    # find query matched by desc so only possibly affected ancestors will be re-queried.
    desc_matching_queries = [
        task_query for task_query in all_task_queries
        if any(
            task_matches_criteria(desc_task, desc_query)
            for desc_query in task_query.get('selectDescendants', []))
    ]
    if desc_matching_queries:
        # The current task matches some descendant criteria. Find all ancestors that match top-level criteria
        # and return those ancestors.
        always_select_fields = get_always_select_fields(all_task_queries)
        for ancestor_task in task_graph.get_ancestors_of_uuid(desc_task.get_uuid()):
            ancestor_query_result = _query_task(
                ancestor_task,
                desc_matching_queries,
                task_graph,
                always_select_fields=always_select_fields
            )
            if ancestor_query_result is not None:
                ancestor_results_by_uuid[ancestor_task.get_uuid()] = ancestor_query_result

    return ancestor_results_by_uuid


def _query_task(
    task: _FlameTask,
    all_task_queries,
    task_graph: FlameTaskGraph,
    always_select_fields=None
):
    matching_queries = [
        query for query in all_task_queries
        if task_matches_criteria(task, query['matchCriteria'])
    ]
    if matching_queries:
        always_select_fields = always_select_fields or get_always_select_fields(all_task_queries)
        select_paths = always_select_fields + flatten([q.get('selectPaths', []) for q in matching_queries])
        all_select_descendants = flatten([q.get('selectDescendants', []) for q in matching_queries])
        return select_from_task(
            select_paths,
            all_select_descendants,
            task,
            task_graph,
        )
    return None


def _select_data_for_matches(task_uuid, task_queries, task_graph: FlameTaskGraph, match_descendant_criteria):
    result_tasks_by_uuid = {}
    task = task_graph._get_task(task_uuid)
    if task is not None:
        query_result = _query_task(task, task_queries, task_graph)
        if query_result is not None:
            result_tasks_by_uuid[task_uuid] = query_result

        if match_descendant_criteria:
            # For incremental updates, find all ancestors with descedant queries that match
            # "task" and query their results too.
            ancestors_by_uuid = _select_ancestors_of_task_descendant_match(
                task,
                task_queries,
                task_graph,
            )
            result_tasks_by_uuid.update(ancestors_by_uuid)

    return result_tasks_by_uuid


def _query_flame_tasks(task_graph: FlameTaskGraph, task_uuids_to_query, task_queries, match_descendant_criteria):
    if not _validate_task_queries(task_queries):
        return {}

    result_tasks_by_uuid = {}
    for uuid in task_uuids_to_query:
        selected_tasks_by_uuid = _select_data_for_matches(uuid, task_queries, task_graph, match_descendant_criteria)
        # Every query result is a full task query, no need for merging. Could de-dupe by senidng "already seen"
        # task uuids to avoid re-querying.
        result_tasks_by_uuid.update(selected_tasks_by_uuid)

    return result_tasks_by_uuid


# Event data extraction/transformation without current state context.
def get_new_event_data(event):
    new_task_data = {}
    for field in COPY_FIELDS:
        if field in event:
            new_task_data[field] = event[field]

    # Note if a field is both a copy field and a transform, the transform overrides if the output writes to the same
    # key.
    for field, transform in FIELD_TO_CELERY_TRANSFORMS.items():
        if field in event:
            new_task_data.update(transform(event))

    return {event['uuid']: new_task_data}

def find_data_changes(task: _FlameTask, new_task_data):
    # Some fields overwrite whatever is present. Be permissive, since not all fields captured are from celery,
    # so not all have entries in the field config.
    override_dict = {
        k: v
        for k, v in new_task_data.items()
        if k not in AGGREGATE_NO_OVERWRITE_FIELDS}

    changed_data = {}
    for new_data_key, new_data_val in override_dict.items():
        if task.get_field(new_data_key) != new_data_val:
            changed_data[new_data_key] = new_data_val

    # Some field updates are dropped if there is already a value for that field name (keep initial).
    for no_overwrite_key in AGGREGATE_KEEP_INITIAL_FIELDS:
        if no_overwrite_key in new_task_data and task.get_field(no_overwrite_key) == _TaskFieldSentile.UNSET:
            changed_data[no_overwrite_key] = new_task_data[no_overwrite_key]

    # Some fields need to be accumulated across events, not overwritten from latest event.
    merged_fields_to_values = {
        field_name: deep_merge(
            task.get_field(
                field_name,
                # default to empty instance of same type
                default=type(new_value)(),
            ),
            new_value,
        )
        for field_name, new_value in new_task_data.items()
        if field_name in AGGREGATE_MERGE_FIELDS
    }
    for merged_data_key, merged_data_val in merged_fields_to_values.items():
        if task.get_field(merged_data_key) != merged_data_val:
            changed_data[merged_data_key] = merged_data_val

    return changed_data


class FlameEventAggregator:
    """ Aggregates many events in to the task data model. """

    def __init__(self, tasks_by_uuid, model_dumper: Optional['FlameModelDumper']=None):
        self.model_dumper = model_dumper
        self._tasks_by_uuid : dict[str, _FlameTask] = tasks_by_uuid
        self.new_task_num : int = len(tasks_by_uuid) + 1

    def aggregate_events(self, events):
        new_data_by_task_uuid = {}
        for e in events:
            event_new_data_by_task_uuid = self._aggregate_event(e)
            for uuid, new_data in event_new_data_by_task_uuid.items():
                if uuid not in new_data_by_task_uuid:
                    new_data_by_task_uuid[uuid] = {}
                new_data_by_task_uuid[uuid].update(new_data)
        return new_data_by_task_uuid

    def generate_incomplete_events(self):
        """
        Unfortunately, if a run terminates ungracefully, incomplete tasks will never arrive at a
        terminal runstate. The 'task-incomplete' runstate is a fake (non-Celery) terminal runstate
        that is generated here so that the UI can show a non-incomplete runstate.
        :return:
        """
        now = datetime.now().timestamp()
        return [
            {
                'uuid': task.get_uuid(),
                'type': RunStates.get_forced_complete_celery_event_type(
                    task.get_field('state'),
                    task.get_field('has_completed') ,
                ),
                'actual_runtime': now - task.get_field('first_started', now)
            }
            for task in self._tasks_by_uuid.values()
            if not task.is_complete()
        ]

    def _get_or_create_task(self, task_uuid) -> tuple[_FlameTask, bool]:
        if task_uuid not in self._tasks_by_uuid:
            task = _FlameTask.create_task(task_uuid, self.new_task_num, self.model_dumper)
            self.new_task_num += 1
            self._tasks_by_uuid[task_uuid] = task
            is_new = True
        else:
            task = self._tasks_by_uuid[task_uuid]
            is_new = False
        return task, is_new

    def _aggregate_event(self, event: dict[str, Any]):
        if (
            # The uuid can be null, it's unclear what this means but the event
            # can't be associated with a task so dropping is OK.
            not event.get('uuid')
            # Revoked events can be sent before any other, and we'll never get any data (name, etc) for that task.
            # Therefore ignore events that are for a new UUID that have revoked type.
            or (
                event['uuid'] not in self._tasks_by_uuid
                and event.get('type') == RunStates.REVOKED.to_celery_event_type()
            )
            # we use task-started-info now
            or event.get('type') == 'task-started'
        ):
            return {}

        new_data_by_task_uuid = get_new_event_data(event)
        changes_by_task_uuid = {}
        for task_uuid, new_task_data in new_data_by_task_uuid.items():
            task, is_new_task = self._get_or_create_task(task_uuid)

            changed_data = find_data_changes(task, new_task_data)
            task.update(changed_data)

            # If we just created the task, we need to send the auto-initialized fields, as well as data from the event.
            # If this isn't a new event, we only need to send what has changed.
            changes_by_task_uuid[task_uuid] = task.as_dict() if is_new_task else dict(changed_data)

        return changes_by_task_uuid


class FlameModelDumper:

    def __init__(self, firex_logs_dir=None, root_model_dir: Optional[str]=None):
        assert bool(firex_logs_dir) ^ bool(root_model_dir), \
            "Dumper needs exclusively either logs dir or root model dir."
        if firex_logs_dir:
            self.root_model_dir = get_flame_model_dir(firex_logs_dir)
        else:
            assert root_model_dir
            self.root_model_dir = root_model_dir
        os.makedirs(self.root_model_dir, exist_ok=True)

        self.full_tasks_dir = get_all_tasks_dir(root_model_dir=self.root_model_dir)
        os.makedirs(self.full_tasks_dir, exist_ok=True)
        self.slim_tasks_file = get_tasks_slim_file(root_model_dir=self.root_model_dir)

    def dump_metadata(self, run_metadata, root_complete, flame_complete):
        metadata_model_file = get_run_metadata_file(root_model_dir=self.root_model_dir)
        complete = {'run_complete': root_complete, 'flame_recv_complete': flame_complete}
        atomic_write_json(metadata_model_file, run_metadata | complete)
        return metadata_model_file

    def _get_full_task_file_path(self, uuid):
        return os.path.join(self.full_tasks_dir, f'{uuid}.json')

    def dump_full_task(self, uuid, task):
        atomic_write_json(self._get_full_task_file_path(uuid), task)

    def load_full_task(self, uuid):
        with open(self._get_full_task_file_path(uuid), encoding='utf-8') as fp:
            return json.load(fp)

    def dump_slim_tasks(self, slim_tasks_by_uuid: dict[str, dict[str, Any]]):
        atomic_write_json(
            self.slim_tasks_file,
            slim_tasks_by_uuid)

    def dump_complete_data_model(
        self,
        task_graph: FlameTaskGraph,
        run_metadata=None,
        dump_task_jsons=True,
    ):
        logger.info("Starting to dump complete Flame model.")

        if dump_task_jsons:
            # Write JSON file with minimum amount of info to render graph.
            self.dump_slim_tasks(task_graph.get_slim_tasks_by_uuid())

            # Write one JSON file per task.
            for uuid, task in task_graph.get_full_tasks_by_uuid().items():
                self.dump_full_task(uuid, task)
            logger.info("Completed dumping Flame tasks.")

        paths_to_compress = [self.slim_tasks_file, self.full_tasks_dir]
        if run_metadata:
            # Write metadata file.
            # Note that since a flame can terminate (e.g. via timeout) before a run, there is no guarantee
            # that the run_metadata model file will ever have root_complete: true.
            root_uuid = task_graph.root_uuid
            root_complete = task_graph.is_root_complete()
            run_metadata_with_root = {**run_metadata, 'root_uuid': root_uuid}
            metadata_model_file = self.dump_metadata(run_metadata_with_root, root_complete, flame_complete=True)
            paths_to_compress.append(metadata_model_file)

        # Write a tar.gz file containing all the files dumped above.
        _dump_full_task_state_archive(self.root_model_dir, paths_to_compress)

        Path(get_model_complete_file(root_model_dir=self.root_model_dir)).touch()
        logger.info("Finished dumping complete Flame model.")

    def dump_task_representation(
        self,
        model_file_name: str,
        tasks_representation: dict[str, Any],
        force=False,
        min_age_change=0,
    ):
        out_file = os.path.join(self.root_model_dir, model_file_name)
        try:
            should_write = (
                force
                or not os.path.exists(out_file)
                or time.time() - os.path.getmtime(out_file) > min_age_change
            )
            if should_write:
                logger.info(f"Starting to dump task representation of: {model_file_name}.")
                atomic_write_json(out_file, tasks_representation)
                logger.info(f"Finished dumping {len(tasks_representation)} task representation of {model_file_name} to {out_file}.")
        except Exception as ex:
            # Don't interfere with shutdown even if extra representation dumping fails.
            logger.error(f"Failed to dump representation of {model_file_name}.")
            logger.exception(ex)


class NoWritngModelDumper(FlameModelDumper):

    def __init__(self, firex_logs_dir):
        super().__init__(firex_logs_dir)

    def dump_full_task(self, *args, **kwargs):
        pass

    def dump_slim_tasks(self, *args, **kwargs):
        pass

    def dump_complete_data_model(self, *args, **kwargs):
        pass

    def dump_task_representation(self, *args, **kwargs):
        pass


def _dump_full_task_state_archive(
    root_model_dir,
    paths_to_compress
):
    logger.info("Starting to create full task state archive.")
    full_state_gz_basename = 'full-run-state.tar.gz'
    tar_bin = shutil.which('tar')
    if tar_bin:
        rel_paths_to_compress = [os.path.relpath(p, root_model_dir) for p in paths_to_compress]
        try:
            subprocess.run(
                [tar_bin, 'czf', full_state_gz_basename] + rel_paths_to_compress,
                cwd=root_model_dir,
                timeout=5*60,
                check=False,
            )
        except subprocess.TimeoutExpired:
            pass
    else:
        with tarfile.open(os.path.join(root_model_dir, full_state_gz_basename), "w:gz") as tar:
            for path in paths_to_compress:
                tar.add(path, arcname=os.path.basename(path))
    logger.info("Completed creating full task state archive.")


"""
Flask API module for interacting with celery tasks.
"""

import logging
from socket import gethostname
import os
import subprocess
import requests
from typing import Optional
import urllib.parse
import getpass

from flask import jsonify, request
from gevent import spawn, sleep
import paramiko

from firexapp.engine.run_controller import FireXRunController
from firex_flame.flame_helper import wait_until, REVOKE_REASON_KEY
from firex_flame.flame_task_graph import FlameTaskGraph, is_task_dict_complete
from firex_flame.controller import FlameAppController
from firex_flame.model_dumper import wait_and_get_flame_url


logger = logging.getLogger(__name__)

subprocess_dict = {}


def _run_metadata_to_api_model(run_metadata, root_uuid):
    return {
        'uid': run_metadata['uid'],
        'logs_dir': run_metadata['logs_dir'],
        'root_uuid': root_uuid,
        'chain': run_metadata['chain'],
        'logs_server': run_metadata['logs_server'],
    }


def poll_channel_readable(channel, timeout=0):
    interval = 0.1
    so_far = 0
    while so_far <= timeout:
        if channel.recv_ready():
            return True
        sleep(interval)
        so_far += interval
    return False


def monitor_file(sio_server, sid, host, filename):
    # To avoid issues with huge log files, we only get the last 50000 lines
    max_lines = 50000

    # helper to send data
    def emit_line_data(data):
        sio_server.emit('file-data', data, room=sid)

    # check if host is localhost or remote - use ssh if remote
    if host in ['127.0.0.1', 'localhost', gethostname()]:
        # Read file locally - output to be sent to requesting client
        logger.info(f"Will start monitoring file {filename} locally")

        if not os.path.isfile(filename):
            emit_line_data(f"File {filename} does not exist.")
            return

        if not os.access(filename, os.R_OK):
            emit_line_data(f"File {filename} is not accessible.")
            return

        try:
            proc = subprocess.Popen(['/usr/bin/tail', '-n', str(max_lines), '--follow=name', filename, '2>/dev/null'],
                                    stdout=subprocess.PIPE)
        except Exception:
            logger.warning("Exception raised while trying to spawn subprocess to monitor file: ", exc_info=True)
            return

        # keep track of sub-procs for later cleanup
        subprocess_dict[sid] = proc

        while True:
            line = proc.stdout.readline()
            if not line:
                break
            emit_line_data(line.decode('utf-8'))

    else:
        # spawn ssh to host to tail -f the file - output to be sent to requesting client
        logger.info("Will start monitoring file %s on host %s" % (filename, host))
        # noinspection PyBroadException
        try:
            ssh = paramiko.SSHClient()
            ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            ssh.connect(host, 22, timeout=60, compress=True)

            # Run find to locate file and match perms
            _, stdout, stderr = ssh.exec_command("\\find %s -perm -004" % filename)
            # Wait for command to return
            stdout.channel.recv_exit_status()

            # Read command stdout/err
            res_out = stdout.read()
            res_err = stderr.read()

            # Check our results
            if not res_out or res_out == b'':
                if not res_err or res_err == b'':
                    emit_line_data("ERROR: File access permissions prevent viewing of file.\n")
                else:
                    res_err = res_err.decode('utf-8', 'ignore')
                    if "No such file or directory" in res_err:
                        # File no longer exists on remote host
                        emit_line_data('[Temporary file no longer exists - executed command has completed]\n')
                    else:
                        emit_line_data("ERROR: Unexpected error while checking file existence and permissions: %s" %
                                       res_err)
                return
            else:
                res_out = res_out.decode('utf-8', 'ignore')
                if filename not in res_out:
                    emit_line_data("ERROR: Unexpected output while checking file existence and permissions: %s" % res_out)
                    return

            try:
                # File exists and has open permissions - tail it
                _, stdout, _ = ssh.exec_command("""/bin/bash -c '/usr/bin/tail -n %d --follow=name %s 2>/dev/null' """ %
                                            (max_lines, filename), bufsize=128, get_pty=True)

                # Keep track of all spawned processes to be able to manage them later
                subprocess_dict[sid] = ssh

                # Set read timeout to 1s
                stdout.channel.settimeout(1)

                # local helper function
                def get_data_chunk():
                    from socket import timeout
                    data_chunk = ''
                    max_chunk_lines = 10000
                    num_chunk_lines = 0
                    end_of_file = False
                    while num_chunk_lines < max_chunk_lines:
                        try:
                            line = stdout.readline()
                            if not line or line == '':
                                # empty line signifies eof
                                end_of_file = True
                                break
                        except timeout:
                            # No more data available within timeout: consider this a full data_chunk to be sent off
                            break
                        else:
                            data_chunk += line
                            num_chunk_lines += 1
                    return data_chunk, num_chunk_lines, end_of_file

                total_num_lines = 0
                eof = False
                while not eof:
                    chunk, num_lines, eof = get_data_chunk()
                    if num_lines:
                        if not total_num_lines:
                            chunk = "[start of data received]\n" + chunk
                        emit_line_data(chunk)
                        total_num_lines += num_lines

                if total_num_lines:
                    emit_line_data('[End of file - program exited]\n')
                    return

            except Exception:
                logger.warning("Exception raised while trying to spawn subprocess to monitor file: ", exc_info=True)

        except Exception as e:
            emit_line_data("ERROR: Spawned subprocess to monitor file failed:\n")
            emit_line_data(str(e))


def term_subproc(sid):
    if sid not in subprocess_dict:
        logger.warning(f"SID {sid} not in subprocess list")
        return

    subproc = subprocess_dict[sid]
    subproc.terminate()
    del subprocess_dict[sid]


def term_all_subprocs():
    for sid in subprocess_dict:
        term_subproc(sid)


def create_socketio_task_api(controller: FlameAppController):

    @controller.sio_server.on('send-graph-state')
    def emit_frontend_tasks_by_uuid(sid, data=None):
        """ Send 'slim' fields for all tasks. This allows visualization of the graph."""
        if (
            isinstance(data, dict)
            and {'task_queries', 'model_file_name'}.intersection(data.keys())
        ):
            tasks_to_send = controller.query_full_tasks(
                data.get('task_queries'),
                data.get('model_file_name'),
            )
        else:
            tasks_to_send = controller.graph.get_slim_tasks_by_uuid()
        controller.sio_server.emit('graph-state', tasks_to_send, room=sid)

    @controller.sio_server.on('send-graph-fields')
    def emit_task_fields_by_uuid(sid, fields):
        """ Send the requested fields for all tasks."""
        response = controller.graph.get_all_tasks_fields(fields)
        controller.sio_server.emit('graph-fields', response, room=sid)

    @controller.sio_server.on('send-run-metadata')
    def emit_run_metadata(sid):
        """ Get static run-level data."""
        response = _run_metadata_to_api_model(
            controller.run_metadata, controller.graph.root_uuid
        )
        controller.sio_server.emit('run-metadata', response, room=sid)

    @controller.sio_server.on('send-task-details')
    def emit_detailed_tasks(sid, uuids):
        """ Get all fields for requested task UUIDs.
        Arguments:
            sid: The session ID to emit to.
            uuids (str or list of str): The uuid of the desired task to get details for.
        """
        if isinstance(uuids, str):
            uuid = uuids
            response = controller.graph.get_full_task_dict(uuid)
            controller.sio_server.emit('task-details-' + uuid, response, room=sid)
        else:
            if not isinstance(uuids, list):
                response = []
            else:
                response = [controller.graph.get_full_task_dict(u) for u in uuids]
            controller.sio_server.emit('task-details', response, room=sid)

    @controller.sio_server.on('start-listen-file')
    def start_file_monitor(sid, args):
        if 'host' not in args or 'filepath' not in args:
            controller.sio_server.emit(
                'file-line',
                f"File monitoring request is missing either host and/or filepath parameters: {args}",
                room=sid)
        else:
            spawn(monitor_file, sio_server=controller.sio_server, sid=sid, host=args['host'], filename=args['filepath'])

    @controller.sio_server.on('stop-listen-file')
    def stop_file_monitor(sid):
        term_subproc(sid)

    @controller.sio_server.on('start-listen-task-query')
    def start_listen_task_query(sid, args):
        if not {'task_queries', 'model_file_name', 'query_config'}.intersection(args.keys()):
            logger.error("Received request to start listening to query without config.")
        else:
            # keep legacy name temporarily for cutover.
            task_queries = args.get('task_queries', args.get('query_config'))
            controller.add_client_task_query_config(
                sid,
                task_queries,
                args.get('model_file_name'),
            )

    @controller.sio_server.on('disconnect')
    def disconnect(sid):
        controller.remove_client_task_query(sid)


def create_rest_task_api(
    web_app,
    task_graph: FlameTaskGraph,
    run_metadata,
):

    @web_app.route('/api/tasks')
    def get_all_tasks_by_uuid():
        return jsonify(task_graph.get_slim_tasks_by_uuid())

    @web_app.route('/api/tasks/<uuid>')
    def get_task_details(uuid):
        task = task_graph.get_full_task_dict(uuid)
        if task is None:
            return '', 404
        return jsonify(task)

    @web_app.route('/api/run-metadata')
    def get_run_metadata():
        return jsonify(_run_metadata_to_api_model(run_metadata, task_graph.root_uuid))


def _data_from_environ_path(environ, path, default):
    if len(path) == 3:
        if path[0] == 'request' and path[1] == 'headers':
            header_key = path[2]
            return  environ.get(f'HTTP_{header_key.upper()}', default)
    return default

def _data_from_request_path(path):
    if len(path) == 3:
        if path[0] == 'request' and path[1] == 'headers':
            header_key = path[2]
            return request.headers.get(header_key)
    return None

def _uuid_and_reason_from_revoke_data(revoke_data):
    if isinstance(revoke_data, str):
        uuid = revoke_data
        revoke_reason = None
    elif isinstance(revoke_data, dict):
        uuid = revoke_data.get('uuid')
        revoke_reason = revoke_data.get(REVOKE_REASON_KEY)
    else:
        uuid = None
        revoke_reason = None
    return uuid, revoke_reason

def create_revoke_api(
    controller: FlameAppController,
    web_app,
    run_controller: FireXRunController,
    authed_user_request_path,
):
    assert controller.sio_server is not None

    sids_to_details = {}
    NO_USER = 'nouser'

    @controller.sio_server.event
    def connect(sid, environ):
        requester = _data_from_environ_path(
            environ, authed_user_request_path, NO_USER)
        sids_to_details[sid] = {
            'requester': requester,
        }

    @controller.sio_server.event
    def disconnect(sid):
        sids_to_details.pop(sid, None)

    @controller.sio_server.on('revoke-task')
    def socketio_revoke_task(sid, revoke_data):
        uuid, revoke_reason = _uuid_and_reason_from_revoke_data(revoke_data)
        if uuid:
            requester = sids_to_details[sid].get('requester', NO_USER)
            revoked = _revoke_task(uuid, 'SocketIO', requester, revoke_reason)
            response_event = 'revoke-success' if revoked else 'revoke-failed'
        else:
            response_event = 'revoke-failed'
        controller.sio_server.emit(response_event, room=sid)

    def _rest_revoke_task(uuid):
        authed_user = _data_from_request_path(authed_user_request_path)
        revoke_reason = request.args.get(REVOKE_REASON_KEY)
        claimed_user = request.args.get('revoking_user')

        user = authed_user or claimed_user or NO_USER
        revoked = _revoke_task(uuid, 'REST', user, revoke_reason)
        return '', 200 if revoked else 500

    @web_app.route('/api/revoke/<uuid>', methods=['GET', 'POST'])
    def rest_revoke_task(uuid):
        return _rest_revoke_task(uuid)

    @web_app.route('/api/revoke', methods=['GET', 'POST'])
    def rest_revoke_root_task():
        # a revoke run quickly after submission
        # might be received before the root task is known,
        # so wait a brief amount of time.
        root_uuid = wait_until(
            controller.graph.is_root_started,
            timeout=60, # startup can be slow.
            sleep_for=0.1,
        )
        root_uuid = controller.graph.root_uuid
        logger.debug(f'Revoking entire run via root task UUID: {root_uuid}')
        if not root_uuid:
            return '', 500
        return _rest_revoke_task(root_uuid)

    def _revoke_task(uuid, type, user, revoke_reason):
        msg = f"Received {type} request to revoke {uuid} from user {user}"
        if revoke_reason:
            msg += f' with reason: {revoke_reason}'
            revoke_reason = revoke_reason[:200] # trim since we'll store this.
        logger.info(msg)
        task = controller.graph.get_slim_task_dict(uuid)
        if task is None:
            return False

        if not is_task_dict_complete(task):
            run_controller.revoke_task(
                uuid,
                revoke_reason,
                revoking_user=user,
                is_root_task=controller.graph.root_uuid == uuid,
            )
        else:
            logger.info(f"Task {uuid} already in terminal state {task['state']}")

        # Wait for the task to become revoked
        revoke_timeout = 10

        revoked = wait_until(controller.graph.was_revoked, timeout=revoke_timeout, sleep_for=0.5, uuid=uuid)
        if not revoked:
            task = controller.graph.get_slim_task_dict(uuid)
            logger.warning(
                f"Failed to revoke  task: waited {revoke_timeout} sec and runstate is currently {task['state']}")
        else:
            logger.debug(f"Successfully revoked task {uuid}.")

        return revoked  # If the task was successfully revoked, return true


def flame_revoke(
    logs_dir : str,
    task_uuid : Optional[str]=None, # None revokes the whole run by revoking the root task.
    revoke_reason: Optional[str]=None,
    revoking_user: Optional[str]=getpass.getuser(),
    timeout=10*60,
) -> Optional[requests.Response]:

    flame_url = wait_and_get_flame_url(firex_logs_dir=logs_dir)
    if not flame_url:
        logger.warning(f"Flame URL not found for {logs_dir}; revoke via Flame will likely fail.")
    else:
        # requesting /api/revoke will revoke the root task, which revoked the entire run.
        url_path = '/api/revoke'
        if task_uuid:
            url_path = os.path.join(url_path, task_uuid)

        url_params = {'revoking_user': revoking_user}
        if revoke_reason:
            url_params[REVOKE_REASON_KEY] = revoke_reason

        return requests.get(
            urllib.parse.urljoin(flame_url, url_path),
            params=url_params,
            timeout=timeout)

    return None


import logging
import os
from threading import Thread

import celery
from firexapp.broker_manager.broker_factory import RedisManager

from firex_flame.controller import FlameAppController
from firex_flame.event_file_processor import process_recording_file
from firex_flame.event_broker_processor import BrokerEventConsumerThread
from firex_flame.flame_helper import wait_until_path_exist, FlameServerConfig

logger = logging.getLogger(__name__)


def create_broker_consumer_thread(broker_consumer_config, controller,
                                  recording_file, shutdown_handler, logs_dir):
    try:
        broker_url = RedisManager.get_broker_url_from_logs_dir(logs_dir)
    except:
        logger.error(f"Failed to load broker URL from logs dir: {logs_dir}")
        raise
    else:
        celery_app = celery.Celery(broker=broker_url, backend=broker_url)
        return BrokerEventConsumerThread(celery_app,
                                         controller,
                                         broker_consumer_config,
                                         recording_file,
                                         shutdown_handler)


def start_flame(server_config: FlameServerConfig, broker_consumer_config, run_metadata, shutdown_handler, wait_for_webserver):
    events_from_recording_file = server_config.recording_file and os.path.isfile(server_config.recording_file)
    controller = FlameAppController(
        run_metadata,
        server_config.extra_task_dump_paths,
        # Don't dump model if we are replaying from a recording file.
        dump_model=not events_from_recording_file,
    )
    if not events_from_recording_file:
        event_recv_thread = create_broker_consumer_thread(
            broker_consumer_config,
            controller,
            server_config.recording_file,
            shutdown_handler,
            run_metadata['logs_dir'])
        celery_app = event_recv_thread.celery_app
    else:
        event_recv_thread = Thread(target=process_recording_file, args=(controller, server_config.recording_file))
        celery_app = None
    event_recv_thread.start()

    # If the client (e.g. main firex process) isn't waiting for the web server and it is waiting for the
    # receiver to be ready, speed up startup by waiting for the receiver to be ready before loading
    # any web modules.
    if (
        not wait_for_webserver
        and not events_from_recording_file
        and broker_consumer_config.receiver_ready_file
    ):
        wait_until_path_exist(broker_consumer_config.receiver_ready_file, sleep_for=0.1)

    # Delaying of importing of all web dependencies is a deliberate startup performance optimization.
    # The broker should be listening for events as quickly as possible.
    from firex_flame.web_app import start_web_server #noqa
    return start_web_server(server_config, controller, celery_app)



import json
import logging
import os
from pathlib import Path
import psutil
import time
import signal
from dataclasses import dataclass
from typing import Optional, List, Union
import hashlib
import platform
import socket

from firexapp.submit.uid import Uid

logger = logging.getLogger(__name__)

DEFAULT_FLAME_TIMEOUT = 60 * 60 * 24 * 2
REVOKE_REASON_KEY = 'revoke_reason'


@dataclass(frozen=True)
class FlameServerConfig:
    webapp_port: int
    serve_logs_dir: bool
    recording_file: Optional[str]
    extra_task_dump_paths: List[str]
    authed_user_request_path: List[str]


def get_flame_redirect_file_path(root_logs_dir):
    return os.path.join(root_logs_dir, 'flame.html')


def get_flame_debug_dir(root_logs_dir):
    return os.path.join(root_logs_dir, Uid.debug_dirname, 'flame')


def get_flame_pid_file_path(root_logs_dir):
    return os.path.join(get_flame_debug_dir(root_logs_dir), 'flame.pid')


def get_flame_pid(root_logs_dir):
    return int(Path(get_flame_pid_file_path(root_logs_dir)).read_text().strip())


def wait_until(predicate, timeout, sleep_for, *args, **kwargs):
    max_time = time.time() + timeout
    while time.time() < max_time:
        pred_result = predicate(*args, **kwargs)
        if pred_result:
            return pred_result
        time.sleep(sleep_for)
    return predicate(*args, **kwargs)


def wait_until_pid_not_exist(pid, timeout=7, sleep_for=1):
    return wait_until(lambda p: not psutil.pid_exists(p), timeout, sleep_for, pid)


def web_request_ok(url):
    import requests
    try:
        return requests.get(url).ok
    except requests.exceptions.ConnectionError:
        return False


def wait_until_web_request_ok(url, timeout=10, sleep_for=1):
    return wait_until(web_request_ok, timeout, sleep_for, url)


def wait_until_path_exist(path, timeout=7, sleep_for=1):
    return wait_until(os.path.exists, timeout, sleep_for, path)


def json_file_fn(json_file_path, fn):
    if not os.path.isfile(json_file_path):
        return False
    try:
        file_data = json.loads(Path(json_file_path).read_text())
    except (json.decoder.JSONDecodeError, OSError):
        return False
    else:
        return fn(file_data)


def get_rec_file(log_dir):
    return os.path.join(get_flame_debug_dir(log_dir), 'flame.rec')


def find_rec_file(log_dir):
    # Formerly was used for backwards compatability, now an alias for get_rec_file
    return get_rec_file(log_dir)

def get_hostname():
    myplatform = platform.system()
    myhostname = socket.gethostname()
    myhostname = f"{myhostname}.local" if myplatform== "Darwin" and not myhostname.endswith("local") else myhostname
    return myhostname


def get_flame_url_from_port(port: int) -> str:
    return f'http://{get_hostname()}:{int(port)}'


class PathNotFoundException(Exception):
    pass


def find(keys, input_dict, raise_error=False):
    result = input_dict
    for key in keys:
        try:
            result = result[key]
        except Exception:
            if raise_error:
                raise PathNotFoundException()
            return None
    return result


def filter_paths(input_dict, paths_to_values):
    results = {}
    for in_key, in_vals in input_dict.items():
        results[in_key] = []
        for in_val in in_vals:
            matches_all = all(to_equal == find(p, in_val) for p, to_equal in paths_to_values.items())
            if matches_all:
                results[in_key].append(in_val)
    return results


def kill_flame(log_dir, sig=signal.SIGKILL, timeout=10):
    flame_pid = get_flame_pid(log_dir)
    kill_and_wait(flame_pid, sig, timeout)
    return flame_pid


def kill_and_wait(pid, sig=signal.SIGKILL, timeout=10):
    if psutil.pid_exists(pid):
        os.kill(pid, sig)
        wait_until_pid_not_exist(pid, timeout=timeout)
    return not psutil.pid_exists(pid)


def create_rel_symlink(existing_path, symlink, target_is_directory=False):
    rel_new_file = os.path.relpath(existing_path, start=os.path.dirname(symlink))
    os.symlink(rel_new_file, symlink, target_is_directory=target_is_directory)


class BrokerConsumerConfig:

    def __init__(self, max_retry_attempts, receiver_ready_file, terminate_on_complete):
        self.max_retry_attempts = max_retry_attempts
        self.receiver_ready_file = receiver_ready_file
        self.terminate_on_complete = terminate_on_complete


def is_json_file(file_path):
    try:
        json.loads(Path(file_path).read_text())
    except json.decoder.JSONDecodeError:
        return False
    else:
        return True


def _both_instance(o1, o2, _type):
    return isinstance(o1, _type) and isinstance(o2, _type)


def deep_merge(container1: Union[dict,list,set], container2: Union[dict,list,set]) -> dict:
    if _both_instance(container1, container2, list):
        # TODO: could deep merge nested dicts.
        return container1 + container2
    elif _both_instance(container1, container2, set):
        return container1.union(container2)
    elif _both_instance(container1, container2, (str, float, int, bool)):
        return container2

    dict1 = container1
    dict2 = container2
    result = dict(dict1)

    for d2_key in dict2:
        if d2_key in dict1:
            v1 = dict1[d2_key]
            v2 = dict2[d2_key]
            if _both_instance(v1, v2, dict):
                result[d2_key] = deep_merge(v1, v2)
            elif _both_instance(v1, v2, list):
                result[d2_key] = v1 + v2
            elif _both_instance(v1, v2, set):
                result[d2_key] = v1.union(v2)
            elif v1 == v2:
                # already the same value in both dicts, take from either.
                result[d2_key] = v1
            else:
                # Both d1 and d2 have entries for d2_key, both entries are not dicts or lists or sets,
                # and the values are not the same. This is a conflict.
                # Overwrite d1's value to simulate dict.update() behaviour.
                result[d2_key] = v2
        else:
            # New key for d1, just add it.
            result[d2_key] = dict2[d2_key]
    return result

def flatten(l):
    return [item for sublist in l for item in sublist]

def get_dict_json_md5(query_config):
    return hashlib.md5(json.dumps(query_config, sort_keys=True).encode('utf-8')).hexdigest()



"""
Process events from Celery in to flame data model.
"""

import logging
from pathlib import Path
import threading
import traceback
import json
import time
from typing import Optional
from io import TextIOWrapper
from datetime import datetime, timezone, timedelta

from celery.events import EventReceiver


from firex_flame.controller import FlameAppController
from firex_flame.flame_helper import BrokerConsumerConfig

logger = logging.getLogger(__name__)


def _get_event_lag(event_timestamp, celery_utcoffset):
    # celery_utcoffset is garbage. It's the UTC offset, but
    # event_timestamp is not in this timezone. It looks like Celery
    # assumes the mc is in UTC, so the real offset is mc_offset_hr + celery_utcoffset
    try:
        now_dt = datetime.now(timezone.utc)
        mc_offset_hr = now_dt.astimezone().utcoffset().total_seconds() / 3600
        real_tz_offset_hr = mc_offset_hr + celery_utcoffset
        event_datetime = datetime.fromtimestamp(
            event_timestamp,
            timezone(timedelta(hours=real_tz_offset_hr))
        )
        return (now_dt - event_datetime.replace(tzinfo=timezone.utc)).total_seconds()
    except Exception:
        return None


def _log_event_received(event, event_count):
    lag_msg = ''
    event_timestamp = event.get('timestamp')

    if event_timestamp:
        lag = _get_event_lag(event_timestamp, event.get('utcoffset', 0))
        if lag is not None:
            lag_msg = f' (lag: {lag:.2f})'

    logger.debug(
        f'Received Celery event number {event_count}'
        f' with task uuid: {event.get("uuid")}{lag_msg}')


class BrokerEventConsumerThread(threading.Thread):
    """Events threading class"""

    def __init__(
        self,
        celery_app,
        flame_controller: FlameAppController,
        config: BrokerConsumerConfig,
        recording_file: str,
        shutdown_handler,
    ):
        threading.Thread.__init__(self, daemon=True)
        self.celery_app = celery_app

        self.open_recording_file : Optional[TextIOWrapper]
        if recording_file:
            self.open_recording_file = open(recording_file, "a", encoding="utf-8")
        else:
            self.open_recording_file = None

        self.flame_controller = flame_controller
        self.max_try_interval = 2**config.max_retry_attempts if config.max_retry_attempts is not None else 32
        self.terminate_on_complete = config.terminate_on_complete
        self.stopped_externally = False
        self.shutdown_handler = shutdown_handler
        self._event_count = 0
        self.celery_event_receiver : Optional[EventReceiver] = None

        self.receiver_ready_file : Optional[Path]

        if config.receiver_ready_file:
            self.receiver_ready_file = Path(config.receiver_ready_file)
            assert not self.receiver_ready_file.exists(), \
                f"Receiver ready file must not already exist: {self.receiver_ready_file}"
        else:
            self.receiver_ready_file = None
        self.is_first_receive = True

    def _create_receiver_ready_file(self):
        if self.receiver_ready_file:
            self.receiver_ready_file.touch()

    def run(self):
        """Listen for events from celery"""
        try:
            self._capture_events()
        finally:
            self._cleanup()

    def _cleanup(self):
        try:
            self.flame_controller.finalize_all_tasks()
            if self.open_recording_file:
                self.open_recording_file.close()
                self.open_recording_file = None
        except Exception as ex:
            logger.error("Failed to cleanup during receiver completion.")
            logger.exception(ex)
        finally:
            logger.info("Completed receiver cleanup.")
            if self.terminate_on_complete and not self.stopped_externally:
                self.shutdown_handler.shutdown("Terminating on completion, as requested by input args.")

    def _capture_events(self):
        try_interval = 1
        while not self.flame_controller.is_root_complete():
            try:
                try_interval *= 2
                with self.celery_app.connection() as conn:
                    conn.ensure_connection(max_retries=1, interval_start=0)
                    self.celery_event_receiver = EventReceiver(
                        conn,
                        handlers={"*": self._safe_on_celery_event},
                        app=self.celery_app)
                    try_interval = 1

                    self.celery_event_receiver.capture(limit=None, timeout=None, wakeup=True)
            except (KeyboardInterrupt, SystemExit) as e:
                self.stopped_externally = True
                self.shutdown_handler.shutdown(str(e))
            except Exception: #noqa
                if self.flame_controller.is_root_complete():
                    logger.info("Root task complete; stopping broker receiver (not entire server).")
                    return
                if self.shutdown_handler.shutdown_received:
                    self.stopped_externally = True
                    logger.info("Shutdown handler received shutdown request; stopping broker receiver.")
                    return
                logger.error(traceback.format_exc())
                if try_interval > self.max_try_interval:
                    logger.warning("Maximum broker retry attempts exceeded, stopping receiver (not entire server)."
                                   " Will no longer retry despite incomplete root task.")
                    return
                logger.debug(f"Try interval {try_interval} secs, still worth retrying.")
                time.sleep(try_interval)

    def _safe_on_celery_event(self, event):
        try:
            self._on_celery_event(event)
        except Exception:
            logger.exception(f'Failed to process Celery event: {event}')

    def _on_celery_event(self, event):
        """Callback function for when an event is received

        Arguments:
            event(dict): The event to aggregate and send downstream.
        """
        if self.is_first_receive:
            # Hopefully this is 'flame-indicate-ready' event
            self._create_receiver_ready_file()
            self.is_first_receive = False

        # Append the event to the recording file if it is specified
        if self.open_recording_file:
            # the rec file is just for the record, doesn't matter
            _safe_dump_json_line(event, self.open_recording_file)

        if self._event_count % 100 == 0:
            if self.open_recording_file:
                _safe_flush(self.open_recording_file) # the rec file is just for the record, doesn't matter
            _log_event_received(event, self._event_count)
        self._event_count += 1


        self.flame_controller.update_graph_and_sio_clients([event])

        if (
            self.flame_controller.is_all_tasks_complete()
            and self.celery_event_receiver
        ):
            logger.info("Stopping Celery event receiver because all tasks are complete.")
            self.celery_event_receiver.should_stop = True


def _safe_dump_json_line(data, open_file: TextIOWrapper):
    try:
        json.dump(data, open_file)
        open_file.write("\n")
    except OSError:
        pass

def _safe_flush(open_file: TextIOWrapper):
    try:
        open_file.flush()
    except OSError:
        pass


import json
import logging
import os
from pathlib import Path
import tempfile
from typing import Optional
from gevent.fileobject import FileObject

from firexapp.common import wait_until
from firex_flame.flame_helper import get_flame_debug_dir

logger = logging.getLogger(__name__)


def atomic_write_json(filename, data):
    filename = os.path.realpath(filename)
    filedir = os.path.dirname(filename)
    with tempfile.NamedTemporaryFile(mode='w',
                                     dir=filedir,
                                     delete=False) as f:
        json.dump(
            data,
            fp=FileObject(f, 'w'), # closes fd
            sort_keys=True,
            indent=2)

        os.chmod(f.name, 0o664)
        os.replace(f.name, filename)


def get_flame_model_dir(firex_logs_dir) -> str:
    return os.path.join(get_flame_debug_dir(firex_logs_dir), 'model')


def get_flame_old_model_dir(firex_logs_dir):
    return os.path.join(firex_logs_dir, 'flame_model')


# Temporarily support both current and old model directory locations.
def find_flame_model_dir(firex_logs_dir):
    model_dir = get_flame_model_dir(firex_logs_dir)
    if os.path.isdir(model_dir):
        return model_dir
    return get_flame_old_model_dir(firex_logs_dir)


def _get_base_model_dir(firex_logs_dir=None, root_model_dir=None):
    assert bool(firex_logs_dir) ^ bool(root_model_dir), \
        "Need exclusively either logs dir or root model dir."
    if firex_logs_dir:
        return find_flame_model_dir(firex_logs_dir)
    else:
        return root_model_dir


def get_all_tasks_dir(firex_logs_dir=None, root_model_dir=None):
    return os.path.join(_get_base_model_dir(firex_logs_dir, root_model_dir), 'full-tasks')


def get_run_metadata_file(firex_logs_dir=None, root_model_dir=None):
    return os.path.join(_get_base_model_dir(firex_logs_dir, root_model_dir), 'run-metadata.json')


def get_run_metadata(firex_logs_dir=None, root_model_dir=None):
    run_metadata_file = get_run_metadata_file(firex_logs_dir, root_model_dir)
    if os.path.isfile(run_metadata_file):
        return json.loads(Path(run_metadata_file).read_text())
    return None


def legacy_is_flame_revoked(firex_logs_dir) -> bool:
    run_metadata = get_run_metadata(firex_logs_dir)
    if run_metadata:
        return bool(run_metadata.get('revoke_timestamp'))
    return False


def _get_flame_url(firex_logs_dir) -> Optional[str]:
    metadata = get_run_metadata(firex_logs_dir)
    if metadata:
        return metadata.get('flame_url')
    return None


def wait_and_get_flame_url(firex_logs_dir: str, timeout=15, sleep_for=0.5) -> Optional[str]:
    return wait_until(
        _get_flame_url,
        timeout=timeout,
        sleep_for=sleep_for,
        firex_logs_dir=firex_logs_dir,
    )


def get_tasks_slim_file(firex_logs_dir=None, root_model_dir=None):
    return os.path.join(_get_base_model_dir(firex_logs_dir, root_model_dir), 'slim-tasks.json')


def get_model_complete_file(firex_logs_dir=None, root_model_dir=None):
    return os.path.join(_get_base_model_dir(firex_logs_dir, root_model_dir), 'model-complete')


def load_slim_tasks(log_dir):
    return json.loads(Path(get_tasks_slim_file(log_dir)).read_text())


def socketio_client_task_queries_request(flame_url, task_queries):
    import socketio
    sio_client = socketio.Client()
    sio_client.connect(flame_url)
    try:
        resp = {'tasks_by_uuid': None}
        @sio_client.on('graph-state')
        def my_event(data):
            resp['tasks_by_uuid'] = data

        sio_client.call('send-graph-state', {'task_queries': task_queries})
        assert resp['tasks_by_uuid'] is not None
    finally:
        sio_client.disconnect()

    return resp['tasks_by_uuid']


def load_task_representation(firex_logs_dir, representation_file, consider_running=False):
    with open(representation_file) as fp:
            representation_data = json.load(fp)

    task_rep_file = os.path.join(_get_base_model_dir(firex_logs_dir=firex_logs_dir),
                                 representation_data['model_file_name'])
    # TODO: this assumes the task representation file is only present once it's complete!
    #   If dumping is moved to occur during event reception, this will need to consider run_metadata.is_complete.
    if os.path.isfile(task_rep_file):
        # If file has already been dumped, just load it.
        with open(task_rep_file) as fp:
            return json.load(fp)
    elif consider_running:
        # If file hasn't been dumped, try to query a running flame server.
        run_metadata_file = get_run_metadata_file(firex_logs_dir=firex_logs_dir)
        if os.path.isfile(run_metadata_file):
            with open(run_metadata_file) as fp:
                run_metadata = json.load(fp)
            return socketio_client_task_queries_request(run_metadata['flame_url'],
                                                        representation_data['task_queries'])

    return None


def get_full_task_path(logs_dir, task_uuid):
    return os.path.join(get_all_tasks_dir(firex_logs_dir=logs_dir), task_uuid + '.json')


def load_full_task(log_dir, task_uuid):
    return json.loads(Path(get_full_task_path(log_dir, task_uuid)).read_text())


def index_tasks_by_names(tasks, names):
    tasks_by_name = {n: [] for n in names}
    for t in tasks:
        if t['name'] in names:
            tasks_by_name[t['name']].append(t)
    return tasks_by_name


def get_model_full_tasks_by_names(log_dir, task_names):
    if isinstance(task_names, str):
        task_names = [task_names]
    tasks_by_uuid = get_full_tasks_by_slim_pred(log_dir, lambda st: st.get('name', None) in task_names)
    return index_tasks_by_names(tasks_by_uuid.values(), task_names)


def get_model_slim_tasks_by_names(log_dir, task_names):
    if isinstance(task_names, str):
        task_names = [task_names]
    tasks_by_uuid = get_slim_tasks_by_slim_pred(
        log_dir,
        lambda st: st.get('name') in task_names)
    return index_tasks_by_names(tasks_by_uuid.values(), task_names)

def get_full_tasks_by_slim_pred(log_dir, slim_predicate):
    return {
        uuid: load_full_task(log_dir, uuid)
        for uuid in get_slim_tasks_by_slim_pred(log_dir, slim_predicate)
    }

def get_slim_tasks_by_slim_pred(log_dir, slim_predicate):
    return {
        uuid: slim_task
        for uuid, slim_task in load_slim_tasks(log_dir).items()
        if slim_predicate(slim_task)
    }


def is_dump_complete(firex_logs_dir):
    return os.path.exists(get_model_complete_file(firex_logs_dir=firex_logs_dir))




import os
import gzip
import json

from firex_flame.model_dumper import find_flame_model_dir, get_model_full_tasks_by_names, \
    index_tasks_by_names, get_full_tasks_by_slim_pred
from firex_flame.flame_task_graph import TASK_ARGS
from firex_flame.flame_helper import find_rec_file, find
from firex_flame.controller import FlameAppController


def load_events_from_rec_file(recording_file):
    assert os.path.isfile(recording_file), f"Recording file doesn't exist: {recording_file}"

    real_rec = os.path.realpath(recording_file)
    if real_rec.endswith('.gz'):
        with gzip.open(real_rec, 'rt', encoding='utf-8') as rec:
            event_lines = rec.readlines()
    else:
        with open(recording_file) as rec:
            event_lines = rec.readlines()

    event_count = 0
    for event_line in event_lines:
        if event_line:
            event =  json.loads(event_line)
            yield event
            event_count += 1
            if event_count % 100 == 0:
                print(f"Processed event {event_count} with uuid {event.get('uuid')}")


def process_recording_file(flame_controller: FlameAppController, recording_file: str):

    for event in load_events_from_rec_file(recording_file):
        flame_controller.update_graph_and_sio_clients([event])

    if flame_controller.is_root_complete():
        # Kludge incomplete runstates that will never become terminal.
        flame_controller.finalize_all_tasks()

    if not flame_controller.run_metadata.get('uid'):
        root_task = flame_controller.graph.get_root_task()
        if root_task:
            flame_controller.run_metadata['uid'] = find([TASK_ARGS, 'uid'], root_task)
            flame_controller.run_metadata['chain'] = find([TASK_ARGS, 'chain'], root_task)


def get_tasks_from_rec_file(log_dir=None, rec_filepath=None, mark_incomplete=False):
    assert bool(log_dir) ^ bool(rec_filepath), "Need exclusively either log directory of rec_file path."
    if not rec_filepath:
        rec_file = find_rec_file(log_dir)
    else:
        rec_file = rec_filepath
    assert os.path.exists(rec_file), f"Recording file not found: {rec_file}"
    flame_controller = FlameAppController({'logs_dir': log_dir})
    process_recording_file(flame_controller, rec_file)

    if mark_incomplete:
        flame_controller.finalize_all_tasks()

    return flame_controller.graph.get_full_tasks_by_uuid(), flame_controller.graph.root_uuid


def get_model_or_rec_full_tasks_by_names(logs_dir, task_names):
    if os.path.isdir(find_flame_model_dir(logs_dir)):
        return get_model_full_tasks_by_names(logs_dir, task_names)

    rec_file = find_rec_file(logs_dir)
    if os.path.isfile(rec_file):
        tasks_by_uuid, _ = get_tasks_from_rec_file(rec_filepath=rec_file)
        return index_tasks_by_names(tasks_by_uuid.values(), task_names)

    raise Exception("Found neither model directory or rec_file, no source of task data in: %s" % logs_dir)


def get_model_or_rec_full_tasks_by_uuids(logs_dir, uuids):
    if os.path.isdir(find_flame_model_dir(logs_dir)):
        return get_full_tasks_by_slim_pred(logs_dir, lambda st: st['uuid'] in uuids)

    if os.path.isdir(logs_dir):
        tasks_by_uuid, _ = get_tasks_from_rec_file(log_dir=logs_dir)
        return {u: t for u, t in tasks_by_uuid.items() if u in uuids}

    raise Exception("Found neither model directory or rec_file, no source of task data in: %s" % logs_dir)




# This file was generated by 'versioneer.py' (0.29) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2026-02-19T05:43:47-0800",
 "dirty": false,
 "error": null,
 "full-revisionid": "3b7d8f9834eaa0362b807df5c1132d2a0bdb7b2a",
 "version": "2.6.107"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



import sys

noop_class = type('noop', (object,), {'iter_entry_points': lambda _: []})
sys.modules["pkg_resources"] = noop_class
sys.modules["celery.events.dispatcher"] = type('noop2', (object,), {'EventDispatcher': noop_class})

from gevent import monkey
monkey.patch_all()

import argparse
import distutils.util
import logging
import os
from pathlib import Path
from gevent import signal
from threading import Timer

from firex_flame.main_app import start_flame
from firex_flame.flame_helper import get_flame_debug_dir, get_flame_pid_file_path, DEFAULT_FLAME_TIMEOUT, \
    BrokerConsumerConfig, get_flame_url_from_port, wait_until, FlameServerConfig, REVOKE_REASON_KEY
# Prevent dependencies from taking module loading hit of pkg_resources.
from firexapp.submit.submit import OptionalBoolean

logger = logging.getLogger(__name__)


def _parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--port', help='Port for starting the web server on', type=int, default=0)
    parser.add_argument('--uid', help='Unique identifier for the represented FireX run.')
    parser.add_argument('--logs_dir', help='Logs directory.', default=None, required=True)
    parser.add_argument('--chain', help='Chain of the run.', default=None)
    # TODO: could validate either rec file exists or broker URL is legible from logs_dir is supplied.
    parser.add_argument('--recording', help='A file containing the recording of celery events.', default=None)
    parser.add_argument('--central_server', help='A central web server from which the UI and logs can be served.',
                        default=None)
    parser.add_argument('--central_server_ui_path', help='Path part of the URL from which the central server serves'
                                                         'the UI. Only has meaning when a central_server is supplied.',
                        default=None)
    parser.add_argument('--logs_server', help='A central web server from which logs can be fetched.',
                        default=None)
    parser.add_argument('--central_documentation_url', help='URL linking to main out-of-app docs.',
                        default='http://www.firexapp.com/')
    parser.add_argument('--firex_bin_path', help='Path to a firex executable.',
                        default=None)
    parser.add_argument('--flame_timeout', help='Maximum lifetime of this service, in seconds', type=int,
                        default=DEFAULT_FLAME_TIMEOUT)
    parser.add_argument('--broker_recv_ready_file', help='File to create immediately before capturing celery events.',
                        default=None)
    parser.add_argument('--broker_max_retry_attempts',
                        help='Number of retry attempts if connection with broker is lost. '
                             'Retries are backed-off exponentially with base 2,'
                             'so a value of 3 here with cause sleeps between retries of '
                             '1 sec, 2 sec, 4 sec, before giving up on retries. Default waits at least a total of '
                             '63 seconds before giving up on retries.',
                        type=int,
                        default=5)
    parser.add_argument('--terminate_on_complete',
                        help='Supply if the Flame server should terminate when the run completes. '
                             'Causes the value of --flame_timeout to be ignored entirely.',
                        default=None, const=True, nargs='?', action=OptionalBoolean)
    parser.add_argument('--extra_task_dump_paths',
                        help='Path to files specifying alternate task dump formats.',
                        type=str, default='')
    parser.add_argument('--authed_user_request_path',
                        help='Dot-delimited path specifying where the web server can find the authed user.',
                        type=str, default='')
    parser.add_argument('--serve_logs_dir',
                        help='Whether or not the Flame web server should make the logs_dir of the run available via '
                             'HTTP(S).',
                        type=lambda x: bool(distutils.util.strtobool(x)),
                        default=True)
    parser.add_argument('--wait_for_webserver',
                        type=lambda x: bool(distutils.util.strtobool(x)),
                        default=True)

    return parser.parse_args()


class ShutdownHandler:

    def __init__(self):
        self.shutdown_received = False
        self.web_server = None
        signal.signal(signal.SIGTERM, self.sigterm_handler)
        signal.signal(signal.SIGINT, self.sigint_handler)
        signal.signal(signal.SIGHUP,
                      lambda _, __: logger.warning("Ignoring SIGHUP signal -- will timeout with broker."))

    def sigterm_handler(self, _, __):
        self.shutdown('SIGTERM detected')

    def sigint_handler(self, _, __):
        self.shutdown('SIGINT detected')

    def timeout_handler(self):
        self.shutdown('timeout exceeded')

    def shutdown(self, reason):
        self.shutdown_received = True
        logging.info("Stopping entire Flame Server for reason: %s" % reason)

        # Avoid race condition where webserver is started immediately after shutdown is called.
        wait_until(lambda: self.web_server is not None, timeout=5, sleep_for=0.5)
        if self.web_server:
            self.web_server.stop()

        # lazy load this module for startup performance.
        from firex_flame.api import term_all_subprocs
        term_all_subprocs()
        logging.shutdown()


def _config_logging(root_logs_dir):
    flame_logs_dir = get_flame_debug_dir(root_logs_dir)
    os.makedirs(flame_logs_dir, exist_ok=True)

    Path(get_flame_pid_file_path(root_logs_dir)).write_text(str(os.getpid()))

    logging.basicConfig(
        filename=os.path.join(flame_logs_dir, 'flame.log'),
        format='[%(asctime)s][%(levelname)s][%(name)s]: %(message)s',
        level=logging.DEBUG,
    )
    # This module is very noisy (logs all data sent), so turn up the level.
    logging.getLogger('engineio.server').setLevel(logging.WARNING)
    logging.getLogger('geventwebsocket.handler').setLevel(logging.WARNING)


def _create_run_metadata(cli_args):
    #FIXME: should be a dataclass.
    return {
        'uid': cli_args.uid,
        'logs_dir': cli_args.logs_dir,
        'central_server': cli_args.central_server,
        'central_server_ui_path': cli_args.central_server_ui_path,
        'chain': cli_args.chain,
        'logs_server': cli_args.logs_server,
        'central_documentation_url': cli_args.central_documentation_url,
        # flame_url is lazy loaded since port is only guaranteed to be known after the web server has started.
        'flame_url': None,
        'firex_bin': cli_args.firex_bin_path,
        'root_uuid': None,
        REVOKE_REASON_KEY: None, # FIXME: delete entirely; moved to run.json
        'revoke_timestamp': None, # FIXME: delete entirely; moved to run.json
    }


def create_broker_processor_config(args):
    return BrokerConsumerConfig(args.broker_max_retry_attempts,
                                args.broker_recv_ready_file,
                                args.terminate_on_complete)


class NoopTimer:
    def start(self):
        pass

    def cancel(self):
        pass


def _create_server_config(args) -> FlameServerConfig:
    if args.extra_task_dump_paths:
        extra_task_dump_paths = args.extra_task_dump_paths.split(',')
    else:
        extra_task_dump_paths = []
    if args.authed_user_request_path:
        authed_user_request_path = args.authed_user_request_path.split('.')
    else:
        authed_user_request_path = []
    return FlameServerConfig(
        webapp_port=args.port,
        serve_logs_dir=args.serve_logs_dir,
        recording_file=args.recording,
        extra_task_dump_paths=extra_task_dump_paths,
        authed_user_request_path=authed_user_request_path
    )

def main():
    shutdown_handler = ShutdownHandler()

    args = _parse_args()
    _config_logging(args.logs_dir)
    t = NoopTimer() if args.terminate_on_complete else Timer(args.flame_timeout, shutdown_handler.timeout_handler)
    try:
        t.start()
        logger.info('Starting Flame Server with args: %s' % args)
        web_server = start_flame(
            _create_server_config(args),
            create_broker_processor_config(args),
            _create_run_metadata(args),
            shutdown_handler,
            args.wait_for_webserver,
        )
        # Allow the shutdown handler to stop the web server before we serve_forever.
        shutdown_handler.web_server = web_server
        print(f"Flame server running on: {get_flame_url_from_port(web_server.server_port)}")
        web_server.serve_forever()
    except Exception as e:
        logger.exception(e)
        shutdown_handler.shutdown(str(e))
    finally:
        t.cancel()
        logger.info("Flame server finished.")


if __name__ == '__main__':
    main()



# Need fastentrypoints to monkey patch setuptools for faster console_scripts
# noinspection PyUnresolvedReferences
import fastentrypoints
from setuptools import setup, find_packages
import versioneer

setup(name='firex_flame',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description='FireX event processor and web server.',
      url='https://github.com/FireXStuff/firex-flame',
      author='Core FireX Team',
      author_email='firex-dev@gmail.com',
      license='BSD-3-Clause',
      packages=find_packages(),
      zip_safe=True,
      install_requires=[
            "Flask==3.1.0",
            "Flask-AutoIndex==0.6.6",
            "Werkzeug==3.1.3",

            # Be very careful changing socketio/engineio versions, as even middle numbers have included breakages in the past.
            # Always test version changes of socketio/engineio against the target UI version of socketio-client.
            "python-socketio==5.3.0",

            'firexapp',
            # Middle UI version number indicates expected server API version.
            # e.g 0.20.x means this server adheres to contract version 0.20 between UI and server.
            "firex_flame_ui<0.30",
            "requests",
            "beautifulsoup4",
            "paramiko",
            "gevent-websocket",
            "gevent",
            "importlib-resources",
            "jsonpath-ng",
            "cryptography==42.0.7",
            "bcrypt==4.1.3", # 4.2.0 causes SIGSEGV
      ],
      package_data={
        'firex_flame': ['templates/*.html'],
      },
      entry_points={
          'console_scripts': ['firex_flame = firex_flame.__main__:main'],
          'firex_tracking_service': ['flame_launcher = firex_flame.launcher:FlameLauncher', ],
      },)



from glob import glob
import json
import logging
import os
import importlib_resources
import urllib.parse
from contextlib import ExitStack
import atexit

import socketio
from gevent import pywsgi
from geventwebsocket.handler import WebSocketHandler

from flask import Flask, redirect, send_from_directory, Response, render_template

from firex_flame.api import create_socketio_task_api, create_revoke_api, create_rest_task_api
from firexapp.submit.reporting import REL_COMPLETION_REPORT_PATH
from firexapp.engine.run_controller import FireXRunController
from firex_flame.flame_helper import FlameServerConfig, get_flame_url_from_port
from firex_flame.controller import FlameAppController


logger = logging.getLogger(__name__)

# This static directory can be used during development to test unreleased versions of the UI.
static_files_dir = os.path.join(os.path.dirname(__file__), "static")

if os.path.isdir(static_files_dir):
    UI_RESOURCE_DIR = static_files_dir
else:
    # Load UI from pypi install case.
    file_manager = ExitStack()
    atexit.register(file_manager.close)
    UI_RESOURCE_DIR = str(file_manager.enter_context(importlib_resources.files('firex_flame_ui')))

REL_UI_RESOURCE_PATH = '/ui'

# Never changes per flame instance, populated on first request.
cached_ui_index_response = None


def _get_asset_rel_path(flame_ui_built_root_dir, asset_suffix) -> str:
    asset_glob =  os.path.join(flame_ui_built_root_dir, 'assets', f'index.*.{asset_suffix}')
    found_assets =  glob(asset_glob)
    assert len(found_assets) == 1, f'Found {len(found_assets)} assets matching glob, require exactly 1. Glob: {asset_glob}'
    return os.path.relpath(found_assets[0], flame_ui_built_root_dir)


class FlameResponse(Response):

    def __init__(self, response, **kwargs):
        # Avoid default octet-stream that causes download instead of display for some log files.
        if kwargs.get('mimetype', None) == 'application/octet-stream':
            kwargs['mimetype'] = 'text/plain'
        super(FlameResponse, self).__init__(response, **kwargs)


def create_ui_index_render_function(central_server, central_server_ui_path):
    def ui_root_render():
        global cached_ui_index_response

        if not cached_ui_index_response:
            rel_js_path = _get_asset_rel_path(UI_RESOURCE_DIR, 'js')
            rel_css_path = _get_asset_rel_path(UI_RESOURCE_DIR, 'css')

            if central_server and central_server_ui_path:
                served_app_js_url = urllib.parse.urljoin(
                    central_server,
                    os.path.join(central_server_ui_path, rel_js_path))

                served_app_css_url = urllib.parse.urljoin(
                    central_server,
                    os.path.join(central_server_ui_path, rel_css_path))
            else:
                # When central valus not supplied, the index.html template produces an html file that references
                # only resources served by this server.
                served_app_js_url = os.path.join(REL_UI_RESOURCE_PATH, rel_js_path)
                served_app_css_url = os.path.join(REL_UI_RESOURCE_PATH, rel_css_path)

            cached_ui_index_response = render_template(
                'index.html',
                served_app_js_url=served_app_js_url,
                served_app_css_url=served_app_css_url,
            )
        return cached_ui_index_response

    return ui_root_render


def ok_json_response(response_data):
    r = Response(response=json.dumps(response_data), status=200, mimetype="application/json")
    r.cache_control.max_age = 2592000
    return r


def create_ui_config(run_metadata):
    return {
        # Tells UI to fetch task data from the origin via Flame's socketio API.
        "access_mode": 'socketio-origin',
        "model_path_template": None,
        "is_central": False,
        "central_server": run_metadata['central_server'],
        "central_server_ui_path": run_metadata['central_server_ui_path'],
        "central_documentation_url": run_metadata['central_documentation_url'],
        "firex_bin": run_metadata['firex_bin'],
        "logs_serving": {
            "serve_mode": "central-webserver",
            "url_format": None,
        },
        'rel_completion_report_path': REL_COMPLETION_REPORT_PATH,
        "linkify_prefixes": ["/auto/", "/ws/"],
    }


def create_web_app(run_metadata, serve_logs_dir):
    web_app = Flask(__name__)
    web_app.response_class = FlameResponse
    web_app.secret_key = os.urandom(24)
    web_app.config['JSONIFY_PRETTYPRINT_REGULAR'] = True

    # Redirect root path to UI.
    web_app.add_url_rule('/', 'flame_ui',
                         # TODO: set cache control.
                         create_ui_index_render_function(run_metadata['central_server'],
                                                         run_metadata['central_server_ui_path']))

    # Serve UI artifacts relatively.
    web_app.add_url_rule(REL_UI_RESOURCE_PATH + '/<path:f>', 'ui_artifacts',
                         lambda f: send_from_directory(UI_RESOURCE_DIR, f))

    # img assets are un-prefixed in build artifacts, so redirect to served img directory.
    web_app.add_url_rule('/img/<path:p>', 'ui_img_artifacts', lambda p: redirect(REL_UI_RESOURCE_PATH + '/img/' + p))

    ui_config = create_ui_config(run_metadata)
    web_app.add_url_rule('/flame-ui-config.json', 'ui_config', lambda: ok_json_response(ui_config))

    # Trivial 'OK' endpoint for testing if the server is up.
    web_app.add_url_rule('/alive', 'alive', lambda: ('', 200))

    # Redirect for old URLs. These links should be updated in emails etc, then this redirect removed.
    web_app.add_url_rule('/root/<uuid>', 'subtree', lambda uuid: redirect('/#/root/' + uuid))

    if serve_logs_dir:
        # Add directory listings and file serve for logs.
        from flask import Blueprint
        from flask_autoindex import AutoIndexBlueprint
        auto_index = Blueprint('auto_index', __name__)
        AutoIndexBlueprint(auto_index, browse_root=run_metadata['logs_dir'])
        web_app.register_blueprint(auto_index, url_prefix=run_metadata['logs_dir'])

    return web_app


def start_web_server(
    server_config: FlameServerConfig,
    controller: FlameAppController,
    celery_app,
) -> pywsgi.WSGIServer:
    web_app = create_web_app(controller.run_metadata, server_config.serve_logs_dir)

    controller.set_sio_server(
        # TODO: parametrize cors_allowed_origins.
        socketio.Server(cors_allowed_origins='*', async_mode='gevent')
    )
    sio_web_app = socketio.WSGIApp(controller.sio_server, web_app)

    create_socketio_task_api(controller)
    create_rest_task_api(web_app, controller.graph, controller.run_metadata)

    server = pywsgi.WSGIServer(
        ('', server_config.webapp_port),
        sio_web_app,
        handler_class=WebSocketHandler,
        spawn=30,
    )
    server.start()  # Need to start() to get port if supplied 0.

    if celery_app:
        # Celery app means this is initial launch, not a replay from a rec file.
        # Can only dump initial metadata now that flame_url is set.
        assert server.server_port, 'Web server port not set after start.'
        controller.dump_updated_metadata(
            {'flame_url': get_flame_url_from_port(server.server_port)},
        )
        create_revoke_api(
            controller,
            web_app,
            FireXRunController(
                celery_app,
                controller.run_metadata['logs_dir']),
            server_config.authed_user_request_path,
        )

    return server



import logging
from typing import Any, Optional
from dataclasses import dataclass, field
from enum import Enum
import json
import copy

import jsonpath_ng
import socketio
from gevent import spawn, sleep
from gevent.queue import JoinableQueue

from firex_flame.flame_helper import get_dict_json_md5
from firex_flame.flame_task_graph import FlameTaskGraph, FlameModelDumper, NoWritngModelDumper


logger = logging.getLogger(__name__)


@dataclass
class _LoadedQueryConfig:

    query_config: dict[str, Any]
    model_file_name: Optional[str]
    md5_hash: str
    latest_full_query_result: dict[str, dict[str, Any]] = None
    listening_client_sids: set[str] = field(default_factory=set)

    @staticmethod
    def create_query_config(query_config, model_file_name=None):
        return _LoadedQueryConfig(
            model_file_name=model_file_name,
            # This caches jsonpath parsing, which is expensive.
            query_config=_convert_json_paths_in_query(query_config),
            md5_hash=get_dict_json_md5(query_config),
        )

    def update_latest_and_clients(
        self,
        sio_server: Optional['socketio.Server'],
        changed_uuids: list[str],
        task_graph: FlameTaskGraph,
    ) -> bool:
        latest = self.query_full_tasks(task_graph)
        updated_partial_query_result = task_graph.query_partial_tasks(
            changed_uuids,
            self.query_config,
        )
        latest.update(updated_partial_query_result)

        if (
            sio_server
            and self.listening_client_sids
            and updated_partial_query_result
        ):
            # send new data to clients listening on this query.
            sio_server.emit(
                'tasks-query-update',
                data=updated_partial_query_result,
                room=self.md5_hash)

        return bool(updated_partial_query_result)

    def add_client(self, sio_server, sid):
        if sid not in self.listening_client_sids:
            self.listening_client_sids.add(sid)
            sio_server.enter_room(sid, room=self.md5_hash)

    def query_full_tasks(self, task_graph: FlameTaskGraph, force=False):
        if self.latest_full_query_result is None or force:
            # recalc and update cached.
            self.latest_full_query_result = task_graph.query_full_tasks(self.query_config)
        return self.latest_full_query_result


@dataclass
class _QueryConfigRegistry:

    loaded_query_configs: list[_LoadedQueryConfig] = field(default_factory=list)

    def _find_config_by_hash(self, query_md5_hash):
        for config in self.loaded_query_configs:
            if config.md5_hash == query_md5_hash:
                return config
        return None

    def _find_config_by_name(self, model_file_name):
        for config in self.loaded_query_configs:
            if config.model_file_name == model_file_name:
                return config
        return None

    def _find_config(self, query_config, model_file_name):
        if model_file_name:
            config = self._find_config_by_name(model_file_name)
            if config:
                return config
        query_md5 = get_dict_json_md5(query_config)
        return self._find_config_by_hash(query_md5)

    def add_query_config(self, sio_server, query_config, model_file_name, sid):
        config = self._find_config(query_config, model_file_name)
        if not config:
            config = _LoadedQueryConfig.create_query_config(query_config, model_file_name)
            self.loaded_query_configs.append(config)

        if sio_server and sid:
            config.add_client(sio_server, sid)

    def remove_listening_client(self, sio_server, sid):
        for config in self.loaded_query_configs:
            if sid in config.listening_client_sids:
                config.listening_client_sids.remove(sid)
                sio_server.leave_room(sid, room=config.md5_hash)

    def update_latest_and_listening_clients(
        self,
        sio_server: Optional['socketio.Server'],
        changed_uuids,
        task_graph: FlameTaskGraph,
    ):
        changed_result_configs = []
        for query_config in self.loaded_query_configs:
            changed = query_config.update_latest_and_clients(
                sio_server,
                changed_uuids,
                task_graph,
            )
            if changed:
                changed_result_configs.append(query_config)
        return changed_result_configs

    def query_full_tasks(self, task_queries, task_graph: FlameTaskGraph, model_file_name, force=False):
        config = self._find_config(task_queries, model_file_name)
        if not config:
            config = _LoadedQueryConfig.create_query_config(task_queries, model_file_name)

        return config.query_full_tasks(task_graph, force=force)


class FlameAppController:

    def __init__(
        self,
        run_metadata: dict[str, Any],
        extra_task_representations=tuple(),
        dump_model=True,
        min_age_repr_dump=60,
    ):
        self.run_metadata = run_metadata
        self.min_age_repr_dump = min_age_repr_dump

        if dump_model:
            self.model_dumper = FlameModelDumper(firex_logs_dir=self.run_metadata['logs_dir'])
        else:
            self.model_dumper = NoWritngModelDumper(self.run_metadata.get('logs_dir'))

        self.graph : FlameTaskGraph = FlameTaskGraph(model_dumper=self.model_dumper)

        self.extra_task_representations = extra_task_representations
        self.query_config_registry = _QueryConfigRegistry()
        self.running_dumper_queue = RunningModelDumper(self)

        # Set after creation as a startup optimization.
        self.sio_server : Optional[socketio.Server] = None

    def update_graph_and_sio_clients(self, events: list[dict[str, Any]]) -> None:
        new_data_by_task_uuid, slim_update_data_by_uuid = self.graph.update_graph_from_celery_events(
            events,
        )
        self._update_slim_listening_sio_clients(slim_update_data_by_uuid)

        self._queue_running_dumper_writes(
            bool(slim_update_data_by_uuid),
            new_data_by_task_uuid,
        )

    def _queue_running_dumper_writes(self, slim_changes: bool, new_data_by_task_uuid):
        if slim_changes:
            self.running_dumper_queue.queue_write_slim()

        self.running_dumper_queue.queue_maybe_write_tasks(
            {
                u: self.graph.get_task_field(u, 'type')
                for u in new_data_by_task_uuid
            })

        self.running_dumper_queue.queue_maybe_write_task_reprs(
            _get_changed_uuids(new_data_by_task_uuid),
        )

    def _update_slim_listening_sio_clients(self, slim_update_data_by_uuid):
        # sio_server can be lazy initialized. Since the event receiving process starts before the
        # web modules are loaded, extremely early events can't be delivered.
        if self.sio_server:
            # Avoid sending events if there aren't fields the downstream cares about.
            if slim_update_data_by_uuid:
                self.sio_server.emit('tasks-update', slim_update_data_by_uuid)

    def dump_updated_metadata(self, update: dict[str, Any]) -> None:
        self.run_metadata.update(update)
        self.model_dumper.dump_metadata(self.run_metadata, root_complete=False, flame_complete=False)

    def _dump_task_query_config_results(
        self,
        query_config: _LoadedQueryConfig,
        force_recalc=False,
    ):
        # only dump named queries specified at launch, not ad-hoc queries from socketio clients.
        if query_config.model_file_name:
            self.model_dumper.dump_task_representation(
                query_config.model_file_name,
                tasks_representation=query_config.query_full_tasks(self.graph, force=force_recalc),
                force=force_recalc,
                min_age_change=self.min_age_repr_dump,
            )

    def dump_extra_task_representations(self, force_recalc=True) -> None:
        self.initialize_task_query_registry()

        for query_config in self.query_config_registry.loaded_query_configs:
            # forcing a full re-calc (includin loading any task that might influence the representation)
            # is is likely overkill, but it makes sure there aren't incremental accumulation errors.
            self._dump_task_query_config_results(query_config, force_recalc=force_recalc)

    def dump_full_task(self, uuid, new_event_types):
        self.graph.dump_full_task(uuid, new_event_types)

    def dump_slim_tasks(self) -> None:
        self.model_dumper.dump_slim_tasks(
            self.graph.get_slim_tasks_by_uuid()
        )

    def update_and_dump_task_representations(self, changed_task_uuids):
        changed_query_result_config = self.query_config_registry.update_latest_and_listening_clients(
            self.sio_server,
            changed_task_uuids,
            self.graph,
        )

        for query_config in changed_query_result_config:
            self._dump_task_query_config_results(query_config)

    def add_client_task_query_config(self, sid, query_config, model_file_name):
        self.query_config_registry.add_query_config(
            self.sio_server,
            query_config,
            model_file_name,
            sid)

    def remove_client_task_query(self, sid):
        self.query_config_registry.remove_listening_client(self.sio_server, sid)

    def initialize_task_query_registry(self):
        if not self.query_config_registry.loaded_query_configs:
            # parsing jsonpaths can be surprisingly slow, so do this lazily instead
            # of at construction time.
            for repr_file in self.extra_task_representations:
                task_repr = load_tasks_representation(repr_file)
                self.add_client_task_query_config(
                    sid=None,
                    query_config=task_repr['task_queries'],
                    model_file_name=task_repr['model_file_name'],
                )

    def set_sio_server(self, sio_server: 'socketio.Server'):
        self.sio_server = sio_server

    def query_full_tasks(self, task_queries, model_file_name, force=False):
        return self.query_config_registry.query_full_tasks(
            task_queries,
            self.graph,
            model_file_name,
            force=force,
        )

    def is_root_complete(self):
        return self.graph.is_root_complete()

    def is_all_tasks_complete(self):
        return self.graph.all_tasks_complete()

    def finalize_all_tasks(self):
        " Mark any incomplete tasks as fake-terminal incomplete state, update clients and dump all task data models."

        _, slim_update_data_by_uuid = self.graph.set_and_dump_any_incomplete_tasks()
        self._update_slim_listening_sio_clients(slim_update_data_by_uuid)
        self.running_dumper_queue.wait_stop()

        self.dump_extra_task_representations()
        self.model_dumper.dump_complete_data_model(
            self.graph,
            run_metadata=self.run_metadata,
            #running model dumper will dump task JSONS, don't do it twice.
            dump_task_jsons=False,
        )


def _get_changed_uuids(new_data_by_task_uuid):
    changed_uuids = list(new_data_by_task_uuid.keys())
    for new_task_data in new_data_by_task_uuid.values():
        if new_task_data.get('parent_id'):
            changed_uuids.append(new_task_data['parent_id'])
        if new_task_data.get('additional_children'):
            changed_uuids.extend(new_task_data['additional_children'])
    return changed_uuids


EXTRA_TASK_REPR_DUMP_DELAY_STEP = 5

class QueueItemType(Enum):
    SLIM_DUMP_TYPE = 'SLIM'
    TASK_DUMP_TYPE = 'TASK'
    STOP_DUMP_TYPE = 'STOP'
    EXTRA_REPR_DUMP_TYPE = 'EXTRA_REPR_DUMP_TYPE'


@dataclass
class _QueueItem:
    item_type : QueueItemType
    task_uuid: Optional[str] = None
    celery_event_type: Optional[str] = None


class RunningModelDumper:
    """
    Maintains a queue and a greenlet for async writing the flame data model.
    Having a dedicated greenlet prevents writing the model to disk
    from blocking the main celery event receiver.
    """

    def __init__(self, flame_controller: FlameAppController):
        self.flame_controller : FlameAppController = flame_controller

        # This is a JoinableQueue just to make testing easier. Clients will wait on the greenlet that processes
        # queue items, not the queue itself.
        self._queue : JoinableQueue[_QueueItem] = JoinableQueue()

        self._consume_queue_greenlet = spawn(self._consume_from_queue)

    def _deduplicate_and_maybe_write_full_tasks(self, task_dump_work_items: list[_QueueItem]) -> None:
        task_uuids_to_event_typess_completed : dict[str, set[str]] = {}
        for task_work_item in task_dump_work_items:
            uuid = task_work_item.task_uuid
            event_type = task_work_item.celery_event_type
            assert uuid is not None, "Must have task UUID for TASK_DUMP_TYPE"
            assert event_type is not None, "Must have event type for TASK_DUMP_TYPE"

            if uuid not in task_uuids_to_event_typess_completed:
                task_uuids_to_event_typess_completed[uuid] = set()
            task_uuids_to_event_typess_completed[uuid].add(event_type)

        for uuid, event_types in task_uuids_to_event_typess_completed.items():
            self.flame_controller.dump_full_task(uuid, event_types)

    def _deduplicate_and_update_task_query_results(self, task_dump_work_items: list[_QueueItem]) -> None:
        changed_uuids = set()
        for task_work_item in task_dump_work_items:
            uuid = task_work_item.task_uuid
            assert uuid is not None, "Must have task UUID for EXTRA_REPR_DUMP_TYPE"
            changed_uuids.add(uuid)

        self.flame_controller.update_and_dump_task_representations(changed_uuids)

    def _get_all_from_queue(self) -> list[_QueueItem]:
        item = self._queue.get()

        # drain queue and process all work items at once, de-duplicating work.
        return [item] + [self._queue.get() for _ in range(len(self._queue))]

    def _consume_from_queue(self) -> None:
        self.flame_controller.initialize_task_query_registry()
        consuming = True
        while consuming:
            # drain queue and process all work items at once, de-duplicating work.
            work_items : list[_QueueItem] = self._get_all_from_queue()
            work_item_types : set[QueueItemType] = {t.item_type for t in work_items}

            try:
                if QueueItemType.SLIM_DUMP_TYPE in work_item_types:
                    self.flame_controller.dump_slim_tasks()

                if QueueItemType.TASK_DUMP_TYPE in work_item_types:
                    self._deduplicate_and_maybe_write_full_tasks(
                        [t for t in work_items if t.item_type == QueueItemType.TASK_DUMP_TYPE],
                    )

                if QueueItemType.EXTRA_REPR_DUMP_TYPE in work_item_types:
                    self._deduplicate_and_update_task_query_results(
                        [t for t in work_items if t.item_type == QueueItemType.EXTRA_REPR_DUMP_TYPE],
                    )

            except Exception as e:
                # TODO: narrow exception handling so that an error in handling of one dump_type doesn't fail others.
                logger.error("Failure while processing task-dumping work queue entry.")
                logger.exception(e)
            finally:
                for _ in range(len(work_items)):
                    self._queue.task_done()

                # Must be last, want to process all other work items before we stop processing all future
                # work items.
                if QueueItemType.STOP_DUMP_TYPE in work_item_types:
                    logger.debug("Stopping in progress model dumper.")
                    consuming = False
                else:
                    # Let other greenlets run, possibly let work accumulate in the queue to allow work de-duplication
                    sleep(0.2)

    def queue_write_slim(self) -> None:
        self._queue.put(_QueueItem(QueueItemType.SLIM_DUMP_TYPE))

    def queue_maybe_write_tasks(self, task_uuids_to_event_types):
        for task_uuid, event_type in task_uuids_to_event_types.items():
            self._queue.put(_QueueItem(QueueItemType.TASK_DUMP_TYPE, task_uuid, event_type))

    def queue_maybe_write_task_reprs(self, task_uuids):
        for task_uuid in task_uuids:
            self._queue.put(_QueueItem(QueueItemType.EXTRA_REPR_DUMP_TYPE, task_uuid))

    def wait_stop(self) -> None:
        self.queue_write_slim()
        self._queue.put(_QueueItem(QueueItemType.STOP_DUMP_TYPE))
        self._consume_queue_greenlet.join() # Wait for queue to drain.


def load_tasks_representation(rep_file):
    with open(rep_file, encoding='utf-8') as fp:
        return json.load(fp)


def _convert_json_paths_in_query(task_queries):
    result = copy.deepcopy(task_queries)
    for query in result:
        if 'selectPaths' in query:
            query['selectPaths'] = [jsonpath_ng.parse(f'$.{p}') for p in query['selectPaths']]
    return result